import sys

import pytest

from cumulusci.tasks.bulkdata.select_utils import (
    OPTIONAL_DEPENDENCIES_AVAILABLE,
    SelectOperationExecutor,
    SelectStrategy,
    add_limit_offset_to_user_filter,
    annoy_post_process,
    calculate_levenshtein_distance,
    determine_field_types,
    find_closest_record,
    levenshtein_distance,
    reorder_records,
    split_and_filter_fields,
    vectorize_records,
)

# Check for pandas availability
try:
    import pandas as pd

    PANDAS_AVAILABLE = True
except ImportError:
    PANDAS_AVAILABLE = False


def test_standard_generate_query_without_filter():
    select_operator = SelectOperationExecutor(SelectStrategy.STANDARD)
    sobject = "Contact"  # Assuming no declaration for this object
    limit = 3
    offset = None
    query, fields = select_operator.select_generate_query(
        sobject=sobject, fields=[], user_filter="", limit=limit, offset=offset
    )

    assert f"LIMIT {limit}" in query
    assert "OFFSET" not in query
    assert fields == ["Id"]


def test_standard_generate_query_with_user_filter():
    select_operator = SelectOperationExecutor(SelectStrategy.STANDARD)
    sobject = "Contact"  # Assuming no declaration for this object
    limit = 3
    offset = None
    user_filter = "WHERE Name IN ('Sample Contact')"
    query, fields = select_operator.select_generate_query(
        sobject=sobject, fields=[], user_filter=user_filter, limit=limit, offset=offset
    )

    assert "WHERE" in query
    assert "Sample Contact" in query
    assert "LIMIT" in query
    assert "OFFSET" not in query
    assert fields == ["Id"]


def test_random_generate_query():
    select_operator = SelectOperationExecutor(SelectStrategy.RANDOM)
    sobject = "Contact"  # Assuming no declaration for this object
    limit = 3
    offset = None
    query, fields = select_operator.select_generate_query(
        sobject=sobject, fields=[], user_filter="", limit=limit, offset=offset
    )

    assert f"LIMIT {limit}" in query
    assert "OFFSET" not in query
    assert fields == ["Id"]


# Test Cases for standard_post_process
def test_standard_post_process_with_records():
    select_operator = SelectOperationExecutor(SelectStrategy.STANDARD)
    records = [["001"], ["002"], ["003"]]
    num_records = 3
    sobject = "Contact"
    selected_records, _, error_message = select_operator.select_post_process(
        load_records=None,
        query_records=records,
        num_records=num_records,
        sobject=sobject,
        weights=[],
        fields=[],
        threshold=None,
    )

    assert error_message is None
    assert len(selected_records) == num_records
    assert all(record["success"] for record in selected_records)
    assert all(record["created"] is False for record in selected_records)
    assert all(record["id"] in ["001", "002", "003"] for record in selected_records)


def test_standard_post_process_with_fewer_records():
    select_operator = SelectOperationExecutor(SelectStrategy.STANDARD)
    records = [["001"]]
    num_records = 3
    sobject = "Opportunity"
    selected_records, _, error_message = select_operator.select_post_process(
        load_records=None,
        query_records=records,
        num_records=num_records,
        sobject=sobject,
        weights=[],
        fields=[],
        threshold=None,
    )

    assert error_message is None
    assert len(selected_records) == num_records
    assert all(record["success"] for record in selected_records)
    assert all(record["created"] is False for record in selected_records)
    # Check if records are repeated to match num_records
    assert selected_records.count({"id": "001", "success": True, "created": False}) == 3


def test_standard_post_process_with_no_records():
    select_operator = SelectOperationExecutor(SelectStrategy.STANDARD)
    records = []
    num_records = 2
    sobject = "Lead"
    selected_records, _, error_message = select_operator.select_post_process(
        load_records=None,
        query_records=records,
        num_records=num_records,
        sobject=sobject,
        weights=[],
        fields=[],
        threshold=None,
    )

    assert selected_records == []
    assert error_message == f"No records found for {sobject} in the target org."


# Test cases for Random Post Process
def test_random_post_process_with_records():
    select_operator = SelectOperationExecutor(SelectStrategy.RANDOM)
    records = [["001"], ["002"], ["003"]]
    num_records = 3
    sobject = "Contact"
    selected_records, _, error_message = select_operator.select_post_process(
        load_records=None,
        query_records=records,
        num_records=num_records,
        sobject=sobject,
        weights=[],
        fields=[],
        threshold=None,
    )

    assert error_message is None
    assert len(selected_records) == num_records
    assert all(record["success"] for record in selected_records)
    assert all(record["created"] is False for record in selected_records)


def test_random_post_process_with_no_records():
    select_operator = SelectOperationExecutor(SelectStrategy.RANDOM)
    records = []
    num_records = 2
    sobject = "Lead"
    selected_records, _, error_message = select_operator.select_post_process(
        load_records=None,
        query_records=records,
        num_records=num_records,
        sobject=sobject,
        weights=[],
        fields=[],
        threshold=None,
    )

    assert selected_records == []
    assert error_message == f"No records found for {sobject} in the target org."


def test_similarity_generate_query_no_nesting():
    select_operator = SelectOperationExecutor(SelectStrategy.SIMILARITY)
    sobject = "Contact"  # Assuming no declaration for this object
    limit = 3
    offset = None
    query, fields = select_operator.select_generate_query(
        sobject, ["Name"], [], limit, offset
    )

    assert fields == ["Id", "Name"]
    assert f"LIMIT {limit}" in query
    assert "OFFSET" not in query


def test_similarity_generate_query_with_nested_fields():
    select_operator = SelectOperationExecutor(SelectStrategy.SIMILARITY)
    sobject = "Event"  # Assuming no declaration for this object
    limit = 3
    offset = None
    fields = [
        "Subject",
        "Who.Contact.Name",
        "Who.Contact.Email",
        "Who.Lead.Name",
        "Who.Lead.Company",
    ]
    query, query_fields = select_operator.select_generate_query(
        sobject, fields, [], limit, offset
    )

    assert "WHERE" not in query  # No WHERE clause should be present
    assert query_fields == [
        "Id",
        "Subject",
        "Who.Contact.Name",
        "Who.Contact.Email",
        "Who.Lead.Name",
        "Who.Lead.Company",
    ]
    assert f"LIMIT {limit}" in query
    assert "TYPEOF Who" in query
    assert "WHEN Contact" in query
    assert "WHEN Lead" in query
    assert "OFFSET" not in query


def test_random_generate_query_with_user_filter():
    select_operator = SelectOperationExecutor(SelectStrategy.SIMILARITY)
    sobject = "Contact"  # Assuming no declaration for this object
    limit = 3
    offset = None
    user_filter = "WHERE Name IN ('Sample Contact')"
    query, fields = select_operator.select_generate_query(
        sobject=sobject,
        fields=["Name"],
        user_filter=user_filter,
        limit=limit,
        offset=offset,
    )

    assert "WHERE" in query
    assert "Sample Contact" in query
    assert "LIMIT" in query
    assert "OFFSET" not in query
    assert fields == ["Id", "Name"]


def test_levenshtein_distance():
    assert levenshtein_distance("kitten", "kitten") == 0  # Identical strings
    assert levenshtein_distance("kitten", "sitten") == 1  # One substitution
    assert levenshtein_distance("kitten", "kitte") == 1  # One deletion
    assert levenshtein_distance("kitten", "sittin") == 2  # Two substitutions
    assert levenshtein_distance("kitten", "dog") == 6  # Completely different strings
    assert levenshtein_distance("kitten", "") == 6  # One string is empty
    assert levenshtein_distance("", "") == 0  # Both strings are empty
    assert levenshtein_distance("Kitten", "kitten") == 1  # Case sensitivity
    assert levenshtein_distance("kit ten", "kitten") == 1  # Strings with spaces
    assert (
        levenshtein_distance("levenshtein", "meilenstein") == 4
    )  # Longer strings with multiple differences


def test_find_closest_record_different_weights():
    load_record = ["hello", "world"]
    query_records = [
        ["record1", "hello", "word"],  # Levenshtein distance = 1
        ["record2", "hullo", "word"],  # Levenshtein distance = 1
        ["record3", "hello", "word"],  # Levenshtein distance = 1
    ]
    weights = [2.0, 0.5]

    # With different weights, the first field will have more impact
    closest_record, _ = find_closest_record(load_record, query_records, weights)
    assert closest_record == [
        "record1",
        "hello",
        "word",
    ], "The closest record should be 'record1'."


def test_find_closest_record_basic():
    load_record = ["hello", "world"]
    query_records = [
        ["record1", "hello", "word"],  # Levenshtein distance = 1
        ["record2", "hullo", "word"],  # Levenshtein distance = 1
        ["record3", "hello", "word"],  # Levenshtein distance = 1
    ]
    weights = [1.0, 1.0]

    closest_record, _ = find_closest_record(load_record, query_records, weights)
    assert closest_record == [
        "record1",
        "hello",
        "word",
    ], "The closest record should be 'record1'."


def test_find_closest_record_multiple_matches():
    load_record = ["cat", "dog"]
    query_records = [
        ["record1", "bat", "dog"],  # Levenshtein distance = 1
        ["record2", "cat", "dog"],  # Levenshtein distance = 0
        ["record3", "dog", "cat"],  # Levenshtein distance = 3
    ]
    weights = [1.0, 1.0]

    closest_record, _ = find_closest_record(load_record, query_records, weights)
    assert closest_record == [
        "record2",
        "cat",
        "dog",
    ], "The closest record should be 'record2'."


def test_similarity_post_process_with_records():
    select_operator = SelectOperationExecutor(SelectStrategy.SIMILARITY)
    num_records = 1
    sobject = "Contact"
    load_records = [["Tom Cruise", "62", "Actor"]]
    query_records = [
        ["001", "Bob Hanks", "62", "Actor"],
        ["002", "Tom Cruise", "63", "Actor"],  # Slight difference
        ["003", "Jennifer Aniston", "30", "Actress"],
    ]

    weights = [1.0, 1.0, 1.0]  # Adjust weights to match your data structure

    selected_records, _, error_message = select_operator.select_post_process(
        load_records=load_records,
        query_records=query_records,
        num_records=num_records,
        sobject=sobject,
        weights=weights,
        fields=["Name", "Age", "Occupation"],
        threshold=None,
    )

    assert error_message is None
    assert len(selected_records) == num_records
    assert all(record["success"] for record in selected_records)
    assert all(record["created"] is False for record in selected_records)
    x = [record["id"] for record in selected_records]
    print(x)
    assert all(record["id"] in ["002"] for record in selected_records)


def test_similarity_post_process_with_no_records():
    select_operator = SelectOperationExecutor(SelectStrategy.SIMILARITY)
    records = []
    num_records = 2
    sobject = "Lead"
    selected_records, _, error_message = select_operator.select_post_process(
        load_records=None,
        query_records=records,
        num_records=num_records,
        sobject=sobject,
        weights=[1, 1, 1],
        fields=[],
        threshold=None,
    )

    assert selected_records == []
    assert error_message == f"No records found for {sobject} in the target org."


def test_similarity_post_process_with_no_records__zero_threshold():
    select_operator = SelectOperationExecutor(SelectStrategy.SIMILARITY)
    load_records = [["Aditya", "Salesforce"], ["Jawad", "Salesforce"]]
    query_records = []
    num_records = 2
    sobject = "Lead"
    (
        selected_records,
        insert_records,
        error_message,
    ) = select_operator.select_post_process(
        load_records=load_records,
        query_records=query_records,
        num_records=num_records,
        sobject=sobject,
        weights=[1, 1, 1],
        fields=["LastName", "Company"],
        threshold=0,
    )

    # Assert that it inserts everything
    assert selected_records == [None, None]
    assert insert_records[0] == ["Aditya", "Salesforce"]
    assert insert_records[1] == ["Jawad", "Salesforce"]
    assert error_message is None


def test_calculate_levenshtein_distance_basic():
    record1 = ["hello", "world"]
    record2 = ["hullo", "word"]
    weights = [1.0, 1.0]

    # Expected distance based on simple Levenshtein distances
    # Levenshtein("hello", "hullo") = 1, Levenshtein("world", "word") = 1
    expected_distance = (1 / 5 * 1.0 + 1 / 5 * 1.0) / 2  # Averaged over two fields

    result = calculate_levenshtein_distance(record1, record2, weights)
    assert result == pytest.approx(
        expected_distance
    ), "Basic distance calculation failed."

    # Empty fields
    record1 = ["hello", ""]
    record2 = ["hullo", ""]
    weights = [1.0, 1.0]

    # Expected distance based on simple Levenshtein distances
    # Levenshtein("hello", "hullo") = 1, Levenshtein("", "") = 0
    expected_distance = (1 / 5 * 1.0 + 0 * 1.0) / 2  # Averaged over two fields

    result = calculate_levenshtein_distance(record1, record2, weights)
    assert result == pytest.approx(
        expected_distance
    ), "Basic distance calculation with empty fields failed."

    # Partial empty fields
    record1 = ["hello", "world"]
    record2 = ["hullo", ""]
    weights = [1.0, 1.0]

    # Expected distance based on simple Levenshtein distances
    # Levenshtein("hello", "hullo") = 1, Levenshtein("world", "") = 5
    expected_distance = (
        1 / 5 * 1.0 + 5 / 5 * 0.05 * 1.0
    ) / 2  # Averaged over two fields

    result = calculate_levenshtein_distance(record1, record2, weights)
    assert result == pytest.approx(
        expected_distance
    ), "Basic distance calculation with partial empty fields failed."


def test_calculate_levenshtein_distance_weighted():
    record1 = ["cat", "dog"]
    record2 = ["bat", "fog"]
    weights = [2.0, 0.5]

    # Levenshtein("cat", "bat") = 1, Levenshtein("dog", "fog") = 1
    expected_distance = (
        1 / 3 * 2.0 + 1 / 3 * 0.5
    ) / 2.5  # Weighted average over two fields

    result = calculate_levenshtein_distance(record1, record2, weights)
    assert result == pytest.approx(
        expected_distance
    ), "Weighted distance calculation failed."


def test_calculate_levenshtein_distance_records_length_doesnt_match():
    record1 = ["cat", "dog", "cow"]
    record2 = ["bat", "fog"]
    weights = [2.0, 0.5]

    with pytest.raises(ValueError) as e:
        calculate_levenshtein_distance(record1, record2, weights)
    assert "Records must have the same number of fields." in str(e.value)


def test_calculate_levenshtein_distance_weights_length_doesnt_match():
    record1 = ["cat", "dog"]
    record2 = ["bat", "fog"]
    weights = [2.0, 0.5, 3.0]

    with pytest.raises(ValueError) as e:
        calculate_levenshtein_distance(record1, record2, weights)
    assert "Records must be same size as fields (weights)." in str(e.value)


@pytest.mark.skipif(
    not PANDAS_AVAILABLE or not OPTIONAL_DEPENDENCIES_AVAILABLE,
    reason="requires optional dependencies for annoy",
)
def test_all_numeric_columns():
    df_db = pd.DataFrame({"A": ["1", "2", "3"], "B": ["4.5", " 5.5", "6.5"]})
    df_query = pd.DataFrame({"A": ["4", "5", ""], "B": ["4.5", "5.5", "6.5"]})
    weights = [0.1, 0.2]
    expected_output = (
        ["A", "B"],  # numerical_features
        [],  # boolean_features
        [],  # categorical_features
        [0.1, 0.2],  # numerical_weights
        [],  # boolean_weights
        [],  # categorical_weights
    )
    assert determine_field_types(df_db, df_query, weights) == expected_output


@pytest.mark.skipif(
    not PANDAS_AVAILABLE or not OPTIONAL_DEPENDENCIES_AVAILABLE,
    reason="requires optional dependencies for annoy",
)
def test_numeric_columns__one_non_numeric():
    df_db = pd.DataFrame({"A": ["1", "2", "3"], "B": ["4.5", "5.5", "6.5"]})
    df_query = pd.DataFrame({"A": ["4", "5", "6"], "B": ["abcd", "5.5", "6.5"]})
    weights = [0.1, 0.2]
    expected_output = (
        ["A"],  # numerical_features
        [],  # boolean_features
        ["B"],  # categorical_features
        [0.1],  # numerical_weights
        [],  # boolean_weights
        [0.2],  # categorical_weights
    )
    assert determine_field_types(df_db, df_query, weights) == expected_output


@pytest.mark.skipif(
    not PANDAS_AVAILABLE or not OPTIONAL_DEPENDENCIES_AVAILABLE,
    reason="requires optional dependencies for annoy",
)
def test_all_boolean_columns():
    df_db = pd.DataFrame(
        {"A": ["true", "false", "true"], "B": ["false", "true", "false"]}
    )
    df_query = pd.DataFrame(
        {"A": ["true", "false", "true"], "B": ["false", "true", "false"]}
    )
    weights = [0.3, 0.4]
    expected_output = (
        [],  # numerical_features
        ["A", "B"],  # boolean_features
        [],  # categorical_features
        [],  # numerical_weights
        [0.3, 0.4],  # boolean_weights
        [],  # categorical_weights
    )
    assert determine_field_types(df_db, df_query, weights) == expected_output


@pytest.mark.skipif(
    not PANDAS_AVAILABLE or not OPTIONAL_DEPENDENCIES_AVAILABLE,
    reason="requires optional dependencies for annoy",
)
def test_all_categorical_columns():
    df_db = pd.DataFrame(
        {"A": ["apple", "banana", "cherry"], "B": ["dog", "cat", "mouse"]}
    )
    df_query = pd.DataFrame(
        {"A": ["banana", "apple", "cherry"], "B": ["cat", "dog", "mouse"]}
    )
    weights = [0.5, 0.6]
    expected_output = (
        [],  # numerical_features
        [],  # boolean_features
        ["A", "B"],  # categorical_features
        [],  # numerical_weights
        [],  # boolean_weights
        [0.5, 0.6],  # categorical_weights
    )
    assert determine_field_types(df_db, df_query, weights) == expected_output


@pytest.mark.skipif(
    not PANDAS_AVAILABLE or not OPTIONAL_DEPENDENCIES_AVAILABLE,
    reason="requires optional dependencies for annoy",
)
def test_mixed_types():
    df_db = pd.DataFrame(
        {
            "A": ["1", "2", "3"],
            "B": ["true", "false", "true"],
            "C": ["apple", "banana", "cherry"],
        }
    )
    df_query = pd.DataFrame(
        {
            "A": ["1", "3", ""],
            "B": ["true", "true", "true"],
            "C": ["apple", "", "3"],
        }
    )
    weights = [0.7, 0.8, 0.9]
    expected_output = (
        ["A"],  # numerical_features
        ["B"],  # boolean_features
        ["C"],  # categorical_features
        [0.7],  # numerical_weights
        [0.8],  # boolean_weights
        [0.9],  # categorical_weights
    )
    assert determine_field_types(df_db, df_query, weights) == expected_output


@pytest.mark.skipif(
    not PANDAS_AVAILABLE or not OPTIONAL_DEPENDENCIES_AVAILABLE,
    reason="requires optional dependencies for annoy",
)
def test_vectorize_records_mixed_numerical_boolean_categorical():
    # Test data with mixed types: numerical and categorical only
    db_records = [["1.0", "true", "apple"], ["2.0", "false", "banana"]]
    query_records = [["1.5", "true", "apple"], ["2.5", "false", "cherry"]]
    weights = [1.0, 1.0, 1.0]  # Equal weights for numerical and categorical columns
    hash_features = 4  # Number of hashing vectorizer features for categorical columns

    final_db_vectors, final_query_vectors = vectorize_records(
        db_records, query_records, hash_features, weights
    )

    # Check the shape of the output vectors
    assert final_db_vectors.shape[0] == len(db_records), "DB vectors row count mismatch"
    assert final_query_vectors.shape[0] == len(
        query_records
    ), "Query vectors row count mismatch"

    # Expected dimensions: numerical (1) + categorical hashed features (4)
    expected_feature_count = 2 + hash_features
    assert (
        final_db_vectors.shape[1] == expected_feature_count
    ), "DB vectors column count mismatch"
    assert (
        final_query_vectors.shape[1] == expected_feature_count
    ), "Query vectors column count mismatch"


@pytest.mark.skipif(
    not PANDAS_AVAILABLE or not OPTIONAL_DEPENDENCIES_AVAILABLE,
    reason="requires optional dependencies for annoy",
)
@pytest.mark.skipif(
    sys.platform == "darwin" and sys.version_info[:2] in [(3, 11), (3, 13)],
    reason="Annoy library has known compatibility issues on macOS with Python 3.11 and 3.13",
)
@pytest.mark.skipif(
    sys.platform == "linux" and sys.version_info[:2] == (3, 12),
    reason="Annoy library has known compatibility issues on Linux with Python 3.12",
)
def test_annoy_post_process():
    # Test data
    load_records = [["Alice", "Engineer"], ["Bob", "Doctor"]]
    query_records = [["q1", "Alice", "Engineer"], ["q2", "Charlie", "Artist"]]
    weights = [1.0, 1.0, 1.0]  # Example weights

    closest_records, insert_records = annoy_post_process(
        load_records=load_records,
        query_records=query_records,
        similarity_weights=weights,
        all_fields=["Name", "Occupation"],
        threshold=None,
    )

    # Assert the closest records
    assert (
        len(closest_records) == 2
    )  # We expect two results (one for each query record)
    assert (
        closest_records[0]["id"] == "q1"
    )  # The first query record should match the first load record

    # No errors expected
    assert not insert_records


@pytest.mark.skipif(
    not PANDAS_AVAILABLE or not OPTIONAL_DEPENDENCIES_AVAILABLE,
    reason="requires optional dependencies for annoy",
)
@pytest.mark.skipif(
    sys.platform == "darwin" and sys.version_info[:2] in [(3, 11), (3, 13)],
    reason="Annoy library has known compatibility issues on macOS with Python 3.11 and 3.13",
)
@pytest.mark.skipif(
    sys.platform == "linux" and sys.version_info[:2] == (3, 12),
    reason="Annoy library has known compatibility issues on Linux with Python 3.12",
)
def test_annoy_post_process__insert_records():
    # Test data
    load_records = [["Alice", "Engineer"], ["Bob", "Doctor"]]
    query_records = [["q1", "Alice", "Engineer"], ["q2", "Charlie", "Artist"]]
    weights = [1.0, 1.0, 1.0]  # Example weights
    threshold = 0.3

    closest_records, insert_records = annoy_post_process(
        load_records=load_records,
        query_records=query_records,
        similarity_weights=weights,
        all_fields=["Name", "Occupation"],
        threshold=threshold,
    )

    # Assert the closest records
    assert len(closest_records) == 2  # We expect two results (one for each load record)

    # Count matches vs insertions
    matches = [record for record in closest_records if record is not None]
    insertions = [record for record in closest_records if record is None]

    # We should have some matches or insertions
    assert len(matches) + len(insertions) == 2

    # Check that matches have the correct structure
    for match in matches:
        assert "id" in match
        assert match["success"] is True
        assert match["created"] is False
        assert match["id"] in ["q1", "q2"]

    # The number of insertions should match the number of None values in closest_records
    assert len(insert_records) == len(insertions)

    # Each insertion record should match the structure expected
    for insert_record in insert_records:
        assert len(insert_record) == 2  # Name and Occupation
        assert insert_record in [["Alice", "Engineer"], ["Bob", "Doctor"]]


@pytest.mark.skipif(
    not PANDAS_AVAILABLE or not OPTIONAL_DEPENDENCIES_AVAILABLE,
    reason="requires optional dependencies for annoy",
)
@pytest.mark.skipif(
    sys.platform == "darwin" and sys.version_info[:2] in [(3, 11), (3, 13)],
    reason="Annoy library has known compatibility issues on macOS with Python 3.11 and 3.13",
)
@pytest.mark.skipif(
    sys.platform == "linux" and sys.version_info[:2] == (3, 12),
    reason="Annoy library has known compatibility issues on Linux with Python 3.12",
)
def test_annoy_post_process__no_query_records():
    # Test data
    load_records = [["Alice", "Engineer"], ["Bob", "Doctor"]]
    query_records = []
    weights = [1.0, 1.0, 1.0]  # Example weights
    threshold = 0.3

    closest_records, insert_records = annoy_post_process(
        load_records=load_records,
        query_records=query_records,
        similarity_weights=weights,
        all_fields=["Name", "Occupation"],
        threshold=threshold,
    )

    # Assert the closest records
    assert len(closest_records) == 2  # We expect two results (both None)
    assert all(rec is None for rec in closest_records)  # Both should be None
    assert insert_records[0] == [
        "Alice",
        "Engineer",
    ]  # The first insert record should match the second load record
    assert insert_records[1] == [
        "Bob",
        "Doctor",
    ]  # The first insert record should match the second load record


@pytest.mark.skipif(
    not PANDAS_AVAILABLE or not OPTIONAL_DEPENDENCIES_AVAILABLE,
    reason="requires optional dependencies for annoy",
)
@pytest.mark.skipif(
    sys.platform == "darwin" and sys.version_info[:2] in [(3, 11), (3, 13)],
    reason="Annoy library has known compatibility issues on macOS with Python 3.11 and 3.13",
)
@pytest.mark.skipif(
    sys.platform == "linux" and sys.version_info[:2] == (3, 12),
    reason="Annoy library has known compatibility issues on Linux with Python 3.12",
)
def test_annoy_post_process__insert_records_with_polymorphic_fields():
    # Test data
    load_records = [
        ["Alice", "Engineer", "Alice_Contact", "abcd1234"],
        ["Bob", "Doctor", "Bob_Contact", "qwer1234"],
    ]
    query_records = [
        ["q1", "Alice", "Engineer", "Alice_Contact"],
        ["q2", "Charlie", "Artist", "Charlie_Contact"],
    ]
    weights = [1.0, 1.0, 1.0, 1.0]  # Example weights
    threshold = 0.3
    all_fields = ["Name", "Occupation", "Contact.Name", "ContactId"]

    closest_records, insert_records = annoy_post_process(
        load_records=load_records,
        query_records=query_records,
        similarity_weights=weights,
        all_fields=all_fields,
        threshold=threshold,
    )

    # Assert the closest records
    assert len(closest_records) == 2  # We expect two results (one for each load record)

    # Count matches vs insertions
    matches = [record for record in closest_records if record is not None]
    insertions = [record for record in closest_records if record is None]

    # We should have some matches or insertions
    assert len(matches) + len(insertions) == 2

    # Check that matches have the correct structure
    for match in matches:
        assert "id" in match
        assert match["success"] is True
        assert match["created"] is False
        assert match["id"] in ["q1", "q2"]

    # The number of insertions should match the number of None values in closest_records
    assert len(insert_records) == len(insertions)

    # Each insertion record should have the polymorphic field filtered out
    # (ContactId should be removed, but Contact.Name lookup field should remain)
    for insert_record in insert_records:
        assert len(insert_record) == 3  # Name, Occupation, ContactId (load fields)
        # Should be one of the original load records but with ContactId field
        assert insert_record in [
            ["Alice", "Engineer", "abcd1234"],
            ["Bob", "Doctor", "qwer1234"],
        ]


@pytest.mark.skipif(
    not PANDAS_AVAILABLE or not OPTIONAL_DEPENDENCIES_AVAILABLE,
    reason="requires optional dependencies for annoy",
)
@pytest.mark.skipif(
    sys.platform == "darwin" and sys.version_info[:2] in [(3, 11), (3, 13)],
    reason="Annoy library has known compatibility issues on macOS with Python 3.11 and 3.13",
)
@pytest.mark.skipif(
    sys.platform == "linux" and sys.version_info[:2] == (3, 12),
    reason="Annoy library has known compatibility issues on Linux with Python 3.12",
)
def test_single_record_match_annoy_post_process():
    # Mock data where only the first query record matches the first load record
    load_records = [["Alice", "Engineer"], ["Bob", "Doctor"]]
    query_records = [["q1", "Alice", "Engineer"]]
    weights = [1.0, 1.0, 1.0]

    closest_records, insert_records = annoy_post_process(
        load_records=load_records,
        query_records=query_records,
        similarity_weights=weights,
        all_fields=["Name", "Occupation"],
        threshold=None,
    )

    # Both the load records should be matched with the only query record we have
    assert len(closest_records) == 2
    assert closest_records[0]["id"] == "q1"
    assert not insert_records


@pytest.mark.parametrize(
    "filter_clause, limit_clause, offset_clause, expected",
    [
        # Test: No existing LIMIT/OFFSET and no new clauses
        ("SELECT * FROM users", None, None, " SELECT * FROM users"),
        # Test: Existing LIMIT and no new limit provided
        ("SELECT * FROM users LIMIT 100", None, None, "SELECT * FROM users LIMIT 100"),
        # Test: Existing OFFSET and no new offset provided
        ("SELECT * FROM users OFFSET 20", None, None, "SELECT * FROM users OFFSET 20"),
        # Test: Existing LIMIT/OFFSET and new clauses provided
        (
            "SELECT * FROM users LIMIT 100 OFFSET 20",
            50,
            10,
            "SELECT * FROM users LIMIT 50 OFFSET 30",
        ),
        # Test: Existing LIMIT, new limit larger than existing (should keep the smaller one)
        ("SELECT * FROM users LIMIT 100", 150, None, "SELECT * FROM users LIMIT 100"),
        # Test: New limit smaller than existing (should use the new one)
        ("SELECT * FROM users LIMIT 100", 50, None, "SELECT * FROM users LIMIT 50"),
        # Test: Existing OFFSET, adding a new offset (should sum the offsets)
        ("SELECT * FROM users OFFSET 20", None, 30, "SELECT * FROM users OFFSET 50"),
        # Test: Existing LIMIT/OFFSET and new values set to None
        (
            "SELECT * FROM users LIMIT 100 OFFSET 20",
            None,
            None,
            "SELECT * FROM users LIMIT 100 OFFSET 20",
        ),
        # Test: Removing existing LIMIT and adding a new one
        ("SELECT * FROM users LIMIT 200", 50, None, "SELECT * FROM users LIMIT 50"),
        # Test: Removing existing OFFSET and adding a new one
        ("SELECT * FROM users OFFSET 40", None, 20, "SELECT * FROM users OFFSET 60"),
        # Edge case: Filter clause with mixed cases
        (
            "SELECT * FROM users LiMiT 100 oFfSeT 20",
            50,
            10,
            "SELECT * FROM users LIMIT 50 OFFSET 30",
        ),
        # Test: Filter clause with trailing/leading spaces
        (
            "   SELECT * FROM users   LIMIT 100   OFFSET 20   ",
            50,
            10,
            "SELECT * FROM users LIMIT 50 OFFSET 30",
        ),
    ],
)
def test_add_limit_offset_to_user_filter(
    filter_clause, limit_clause, offset_clause, expected
):
    result = add_limit_offset_to_user_filter(filter_clause, limit_clause, offset_clause)
    assert result.strip() == expected.strip()


def test_reorder_records_basic_reordering():
    records = [
        ["Alice", 30, "Engineer"],
        ["Bob", 25, "Designer"],
    ]
    original_fields = ["name", "age", "job"]
    new_fields = ["job", "name"]

    expected = [
        ["Engineer", "Alice"],
        ["Designer", "Bob"],
    ]
    result = reorder_records(records, original_fields, new_fields)
    assert result == expected


def test_reorder_records_partial_fields():
    records = [
        ["Alice", 30, "Engineer"],
        ["Bob", 25, "Designer"],
    ]
    original_fields = ["name", "age", "job"]
    new_fields = ["age"]

    expected = [
        [30],
        [25],
    ]
    result = reorder_records(records, original_fields, new_fields)
    assert result == expected


def test_reorder_records_missing_fields_in_new_fields():
    records = [
        ["Alice", 30, "Engineer"],
        ["Bob", 25, "Designer"],
    ]
    original_fields = ["name", "age", "job"]
    new_fields = ["nonexistent", "job"]

    expected = [
        ["Engineer"],
        ["Designer"],
    ]
    result = reorder_records(records, original_fields, new_fields)
    assert result == expected


def test_reorder_records_empty_records():
    records = []
    original_fields = ["name", "age", "job"]
    new_fields = ["job", "name"]

    expected = []
    result = reorder_records(records, original_fields, new_fields)
    assert result == expected


def test_reorder_records_empty_new_fields():
    records = [
        ["Alice", 30, "Engineer"],
        ["Bob", 25, "Designer"],
    ]
    original_fields = ["name", "age", "job"]
    new_fields = []

    expected = [
        [],
        [],
    ]
    result = reorder_records(records, original_fields, new_fields)
    assert result == expected


def test_reorder_records_empty_original_fields():
    records = [
        ["Alice", 30, "Engineer"],
        ["Bob", 25, "Designer"],
    ]
    original_fields = []
    new_fields = ["job", "name"]

    with pytest.raises(KeyError):
        reorder_records(records, original_fields, new_fields)


def test_reorder_records_no_common_fields():
    records = [
        ["Alice", 30, "Engineer"],
        ["Bob", 25, "Designer"],
    ]
    original_fields = ["name", "age", "job"]
    new_fields = ["nonexistent_field"]

    expected = [
        [],
        [],
    ]
    result = reorder_records(records, original_fields, new_fields)
    assert result == expected


def test_reorder_records_duplicate_fields_in_new_fields():
    records = [
        ["Alice", 30, "Engineer"],
        ["Bob", 25, "Designer"],
    ]
    original_fields = ["name", "age", "job"]
    new_fields = ["job", "job", "name"]

    expected = [
        ["Engineer", "Engineer", "Alice"],
        ["Designer", "Designer", "Bob"],
    ]
    result = reorder_records(records, original_fields, new_fields)
    assert result == expected


def test_reorder_records_all_fields_in_order():
    records = [
        ["Alice", 30, "Engineer"],
        ["Bob", 25, "Designer"],
    ]
    original_fields = ["name", "age", "job"]
    new_fields = ["name", "age", "job"]

    expected = [
        ["Alice", 30, "Engineer"],
        ["Bob", 25, "Designer"],
    ]
    result = reorder_records(records, original_fields, new_fields)
    assert result == expected


def test_split_and_filter_fields_basic_case():
    fields = [
        "Account.Name",
        "Account.Industry",
        "Contact.Name",
        "AccountId",
        "ContactId",
        "CreatedDate",
    ]
    load_fields, select_fields = split_and_filter_fields(fields)
    assert load_fields == ["AccountId", "ContactId", "CreatedDate"]
    assert select_fields == [
        "Account.Name",
        "Account.Industry",
        "Contact.Name",
        "CreatedDate",
    ]


def test_split_and_filter_fields_all_non_lookup_fields():
    fields = ["Name", "CreatedDate"]
    load_fields, select_fields = split_and_filter_fields(fields)
    assert load_fields == ["Name", "CreatedDate"]
    assert select_fields == fields


def test_split_and_filter_fields_all_lookup_fields():
    fields = ["Account.Name", "Account.Industry", "Contact.Name"]
    load_fields, select_fields = split_and_filter_fields(fields)
    assert load_fields == []
    assert select_fields == fields


def test_split_and_filter_fields_empty_fields():
    fields = []
    load_fields, select_fields = split_and_filter_fields(fields)
    assert load_fields == []
    assert select_fields == []


def test_split_and_filter_fields_single_non_lookup_field():
    fields = ["Id"]
    load_fields, select_fields = split_and_filter_fields(fields)
    assert load_fields == ["Id"]
    assert select_fields == ["Id"]


def test_split_and_filter_fields_single_lookup_field():
    fields = ["Account.Name"]
    load_fields, select_fields = split_and_filter_fields(fields)
    assert load_fields == []
    assert select_fields == ["Account.Name"]


def test_split_and_filter_fields_multiple_unique_lookups():
    fields = [
        "Account.Name",
        "Account.Industry",
        "Contact.Email",
        "Contact.Phone",
        "Id",
    ]
    load_fields, select_fields = split_and_filter_fields(fields)
    assert load_fields == ["Id"]
    assert (
        select_fields == fields
    )  # No filtering applied since all components are unique



import contextlib
import functools
import json
import os
import re
import time
from collections import defaultdict

from cumulusci.core.config import BaseProjectConfig, ScratchOrgConfig, TaskConfig
from cumulusci.core.exceptions import ProjectConfigNotFound
from cumulusci.core.sfdx import sfdx
from cumulusci.core.utils import process_bool_arg, process_list_arg
from cumulusci.tasks.metadata.package import PackageXmlGenerator
from cumulusci.tasks.salesforce import BaseRetrieveMetadata, BaseSalesforceApiTask
from cumulusci.tasks.salesforce.retrieve_profile import RetrieveProfile
from cumulusci.utils import (
    inject_namespace,
    process_text_in_directory,
    temporary_dir,
    tokenize_namespace,
    touch,
)

# These types are leftover from core but SFDX does not know how to process them.
# They should be excluded from package manifests files.
KNOWN_BAD_MD_TYPES = [
    "AuraDefinition",
    "ExperienceResource",
    "LightningComponentResource",
]


class ListChanges(BaseSalesforceApiTask):
    api_version = "48.0"

    task_options = {
        "include": {
            "description": "A comma-separated list of strings. "
            "Components will be included if one of these strings "
            "is part of either the metadata type or name. "
            "Example: ``-o include CustomField,Admin`` matches both "
            "``CustomField: Favorite_Color__c`` and ``Profile: Admin``"
        },
        "types": {
            "description": "A comma-separated list of metadata types to include."
        },
        "exclude": {"description": "Exclude changed components matching this string."},
        "snapshot": {
            "description": "If True, all matching items will be set to be ignored at their current revision number.  This will exclude them from the results unless a new edit is made."
        },
    }

    def _init_options(self, kwargs):
        super(ListChanges, self)._init_options(kwargs)
        self.options["include"] = process_list_arg(self.options.get("include", [])) + [
            f"{mdtype}:" for mdtype in process_list_arg(self.options.get("types", []))
        ]
        self.options["exclude"] = process_list_arg(self.options.get("exclude", []))
        self.options["snapshot"] = process_bool_arg(self.options.get("snapshot", False))
        self._include = self.options["include"]
        self._exclude = self.options["exclude"]
        self._exclude.extend(self.project_config.project__source__ignore or [])

    @property
    @contextlib.contextmanager
    def _snapshot_file(self):
        with self.project_config.open_cache("snapshot") as parent_dir:
            yield parent_dir / f"{self.org_config.name}.json"

    def _load_snapshot(self):
        """Load the snapshot of which component revisions have been retrieved."""
        self._snapshot = {}
        with self._snapshot_file as sf:
            if sf.exists():
                with sf.open("r", encoding="utf-8") as f:
                    self._snapshot = json.load(f)

    def _run_task(self):
        self._load_snapshot()
        changes = self._get_changes()
        if changes:
            self.logger.info(
                f"Found {len(changes)} changed components in the scratch org."
            )
        else:
            self.logger.info("Found no changes.")

        filtered, ignored = self._filter_changes(changes)
        if ignored:
            self.logger.info(
                f"Ignored {len(ignored)} changed components in the scratch org."
            )
            self.logger.info(f"{len(filtered)} remaining changes after filtering.")

        for change in filtered:
            self.logger.info("{MemberType}: {MemberName}".format(**change))

        if self.options["snapshot"]:
            self.logger.info("Storing snapshot of changes")
            self._store_snapshot(filtered)

    def _get_changes(self):
        """Get the SourceMember records that have changed since the last snapshot."""
        sourcemembers = self.tooling.query_all(
            "SELECT MemberName, MemberType, RevisionCounter FROM SourceMember "
            "WHERE IsNameObsolete=false"
        )
        changes = []
        for sourcemember in sourcemembers["records"]:
            mdtype = sourcemember["MemberType"]
            name = sourcemember["MemberName"]
            current_revnum = self._snapshot.get(mdtype, {}).get(name)
            new_revnum = sourcemember["RevisionCounter"] or -1
            if current_revnum and current_revnum == new_revnum:
                continue
            changes.append(sourcemember)
        return changes

    def _filter_changes(self, changes):
        """Filter changes using the include/exclude options"""
        filtered = []
        ignored = []
        for change in changes:
            mdtype = change["MemberType"]
            name = change["MemberName"]
            full_name = f"{mdtype}: {name}"
            if (
                self._include
                and not any(re.search(s, full_name) for s in self._include)
            ) or any(re.search(s, full_name) for s in self._exclude):
                ignored.append(change)
            else:
                filtered.append(change)
        return filtered, ignored

    def _store_snapshot(self, changes):
        """Update the snapshot of which component revisions have been retrieved."""
        for change in changes:
            mdtype = change["MemberType"]
            name = change["MemberName"]
            revnum = change["RevisionCounter"] or -1
            self._snapshot.setdefault(mdtype, {})[name] = revnum
        with self._snapshot_file as sf:
            with sf.open("w", encoding="utf-8") as f:
                json.dump(self._snapshot, f)

    def _reset_sfdx_snapshot(self):
        # If org is from sfdx, reset sfdx source tracking
        if self.project_config.project__source_format == "sfdx" and isinstance(
            self.org_config, ScratchOrgConfig
        ):
            sfdx(
                "project reset tracking",
                args=["-p"],
                username=self.org_config.username,
                capture_output=True,
                check_return=True,
            )


retrieve_changes_task_options = ListChanges.task_options.copy()
retrieve_changes_task_options["path"] = {
    "description": "The path to write the retrieved metadata",
    "required": False,
}
retrieve_changes_task_options["api_version"] = {
    "description": (
        "Override the default api version for the retrieve."
        + " Defaults to project__package__api_version"
    )
}
retrieve_changes_task_options["retrieve_complete_profile"] = {
    "description": (
        "If set to True, will use RetrieveProfile to retrieve"
        + " the complete profile. Default is set to False"
    )
}
retrieve_changes_task_options["namespace_tokenize"] = BaseRetrieveMetadata.task_options[
    "namespace_tokenize"
]

retrieve_changes_task_options["output_dir"] = {
    "description": (
        "The output directory for the retrieved metadata. "
        + "If not specified, defaults to force-app or the target directory passed to retrieve changes."
    ),
    "required": False,
}


def _write_manifest(changes, path, api_version):
    """Write a package.xml for the specified changes and API version."""
    type_members = defaultdict(list)
    for change in changes:
        mdtype = change["MemberType"]
        if mdtype in KNOWN_BAD_MD_TYPES:
            continue
        # folders are retrieved along with their contained type
        if mdtype.endswith("Folder"):
            mdtype = mdtype[: -len("Folder")]
        type_members[mdtype].append(change["MemberName"])

    generator = PackageXmlGenerator(
        ".",
        api_version,
        types=[MetadataType(name, members) for name, members in type_members.items()],
    )
    package_xml = generator()
    with open(os.path.join(path, "package.xml"), "w", encoding="utf-8") as f:
        f.write(package_xml)


def separate_profiles(components):
    """Separate the profiles from components"""
    updated_components = []
    profiles = []
    for comp in components:
        if comp["MemberType"] == "Profile":
            profiles.append(comp["MemberName"])
        else:
            updated_components.append(comp)

    return updated_components, profiles


def retrieve_components(
    components,
    org_config,
    target: str,
    md_format: bool,
    extra_package_xml_opts: dict,
    namespace_tokenize: str,
    api_version: str,
    project_config: BaseProjectConfig = None,
    retrieve_complete_profile: bool = False,
    capture_output: bool = False,
    output_dir: str = None,
):
    """Retrieve specified components from an org into a target folder.

    Retrieval is done using the sf project retrieve start command.

    Set `md_format` to True if retrieving into a folder with a package
    in metadata format. In this case the folder will be temporarily
    converted to dx format for the retrieval and then converted back.
    Retrievals to metadata format can also set `namespace_tokenize`
    to a namespace prefix to replace it with a `%%%NAMESPACE%%%` token.
    """

    # Always use output_dir if specified, else use target
    retrieve_target = (
        os.path.realpath(output_dir) if output_dir else os.path.realpath(target)
    )
    profiles = []
    # If retrieve_complete_profile and project_config is None, raise error
    # This is because project_config is only required if retrieve_complete_profile is True
    if retrieve_complete_profile and project_config is None:
        raise ProjectConfigNotFound(
            "Kindly provide project_config as part of retrieve_components"
        )

    with contextlib.ExitStack() as stack:
        if md_format:
            # Create target if it doesn't exist
            if not os.path.exists(target):
                os.mkdir(target)
                touch(os.path.join(target, "package.xml"))

            # Inject namespace
            if namespace_tokenize:
                process_text_in_directory(
                    target,
                    functools.partial(
                        inject_namespace, namespace=namespace_tokenize, managed=True
                    ),
                )

            # Temporarily convert metadata format to DX format
            stack.enter_context(temporary_dir())
            os.mkdir("target")
            # We need to create sfdx-project.json
            # so that sfdx will recognize force-app as a package directory.
            with open("sfdx-project.json", "w", encoding="utf-8") as f:
                json.dump(
                    {"packageDirectories": [{"path": "force-app", "default": True}]}, f
                )
            sfdx(
                "project convert mdapi",
                log_note="Converting to DX format",
                args=["-r", target, "-d", "force-app"],
                check_return=True,
            )

        # If retrieve_complete_profile is True, separate the profiles from
        # components to retrieve complete profile
        if retrieve_complete_profile:
            components, profiles = separate_profiles(components)

        if components:
            # Construct package.xml with components to retrieve, in its own tempdir
            package_xml_path = stack.enter_context(temporary_dir(chdir=False))
            _write_manifest(components, package_xml_path, api_version)

            # Build args list conditionally including --output-dir
            # Use relative path for package.xml to avoid SF CLI path issues
            sfdx_args = [
                "-a",
                str(api_version),
                "-x",
                os.path.join(package_xml_path, "package.xml"),
                "-w",
                "5",
                "--ignore-conflicts",
            ]

            # Only add --output-dir if output_dir was specified
            if output_dir:
                sfdx_args.extend(["--output-dir", retrieve_target])

            # Retrieve specified components in DX format
            p = sfdx(
                "project retrieve start",
                access_token=org_config.access_token,
                log_note="Retrieving components",
                args=sfdx_args,
                capture_output=capture_output,
                check_return=True,
                env={"SF_ORG_INSTANCE_URL": org_config.instance_url},
            )

        # Extract Profiles
        if profiles:
            task_config = TaskConfig(
                config={
                    "options": {"profiles": ",".join(profiles), "path": "force-app"}
                }
            )
            cls_retrieve_profile = RetrieveProfile(
                org_config=org_config,
                project_config=project_config,
                task_config=task_config,
            )
            cls_retrieve_profile()
        if md_format:
            # Convert back to metadata format
            sfdx(
                "project convert source",
                log_note="Converting back to metadata format",
                args=["-r", "force-app", "-d", target],
                capture_output=capture_output,
                check_return=True,
            )

            # Reinject namespace tokens
            if namespace_tokenize:
                process_text_in_directory(
                    target,
                    functools.partial(tokenize_namespace, namespace=namespace_tokenize),
                )

            # Regenerate package.xml,
            # to avoid reformatting or losing package name/scripts
            package_xml_opts = {
                "directory": target,
                "api_version": api_version,
                **extra_package_xml_opts,
            }
            package_xml = PackageXmlGenerator(**package_xml_opts)()
            with open(os.path.join(target, "package.xml"), "w", encoding="utf-8") as f:
                f.write(package_xml)
        if capture_output:
            return p.stdout_text.read()
        else:
            return None


class RetrieveChanges(ListChanges, BaseSalesforceApiTask):
    task_options = retrieve_changes_task_options

    def _init_options(self, kwargs):
        super(RetrieveChanges, self)._init_options(kwargs)
        self.options["snapshot"] = process_bool_arg(kwargs.get("snapshot", True))
        self.options["retrieve_complete_profile"] = process_bool_arg(
            self.options.get("retrieve_complete_profile", False)
        )

        # Get output_dir first
        output_dir = self.options.get("output_dir")

        # Check which directories are configured as dx packages
        package_directories = []
        default_package_directory = None
        if os.path.exists("sfdx-project.json"):
            with open("sfdx-project.json", "r", encoding="utf-8") as f:
                sfdx_project = json.load(f)
                for package_directory in sfdx_project.get("packageDirectories", []):
                    package_directories.append(package_directory["path"])
                    if package_directory.get("default"):
                        default_package_directory = package_directory["path"]

        path = self.options.get("path")
        if path is None:
            # set default path to src for mdapi format,
            # or the default package directory from sfdx-project.json for dx format
            if (
                default_package_directory
                and self.project_config.project__source_format == "sfdx"
            ):
                path = default_package_directory
                md_format = False
            else:
                path = "src"
                md_format = True
        else:
            md_format = path not in package_directories
        self.md_format = md_format
        self.options["path"] = path
        self.options["output_dir"] = output_dir

        if "api_version" not in self.options:
            self.options[
                "api_version"
            ] = self.project_config.project__package__api_version

    def _run_task(self):
        self._load_snapshot()
        self.logger.info("Querying Salesforce for changed source members")
        changes = self._get_changes()
        filtered, ignored = self._filter_changes(changes)
        if not filtered:
            self.logger.info("No changes to retrieve")
            return
        for change in filtered:
            self.logger.info("{MemberType}: {MemberName}".format(**change))

        target = os.path.realpath(self.options["path"])
        output_dir = self.options.get("output_dir")
        package_xml_opts = {}
        if self.options["path"] == "src":
            package_xml_opts.update(
                {
                    "package_name": self.project_config.project__package__name,
                    "install_class": self.project_config.project__package__install_class,
                    "uninstall_class": self.project_config.project__package__uninstall_class,
                }
            )

        retrieve_components(
            filtered,
            self.org_config,
            target,
            md_format=self.md_format,
            namespace_tokenize=self.options.get("namespace_tokenize"),
            api_version=self.options["api_version"],
            extra_package_xml_opts=package_xml_opts,
            project_config=self.project_config,
            retrieve_complete_profile=self.options["retrieve_complete_profile"],
            output_dir=output_dir,
        )

        if self.options["snapshot"]:
            self.logger.info("Storing snapshot of changes")
            self._store_snapshot(filtered)

            if not ignored:
                # If all changed components were retrieved,
                # we can reset sfdx source tracking too
                self._reset_sfdx_snapshot()


class SnapshotChanges(ListChanges):

    task_options = {}

    def _init_options(self, kwargs):
        # Avoid loading ListChanges options
        pass

    def _run_task(self):
        if self.org_config.scratch:
            self._snapshot = {}

            changes = self._get_changes()
            if not changes:
                # Try again if source tracking hasn't updated
                time.sleep(5)
                changes = self._get_changes()

            if changes:
                self._store_snapshot(changes)
            self._reset_sfdx_snapshot()

    def freeze(self, step):
        return []


class MetadataType(object):
    def __init__(self, name, members):
        self.metadata_type = name
        self.members = members

    def __call__(self):
        return (
            ["    <types>"]
            + [f"        <members>{member}</members>" for member in self.members]
            + [f"        <name>{self.metadata_type}</name>", "    </types>"]
        )



import re
import time

from Browser import SupportedBrowsers
from Browser.utils.data_types import KeyAction, PageLoadStates
from robot.utils import timestr_to_secs

from cumulusci.robotframework.base_library import BaseLibrary
from cumulusci.robotframework.faker_mixin import FakerMixin
from cumulusci.robotframework.utils import (
    WAIT_FOR_AURA_SCRIPT,
    capture_screenshot_on_error,
)


class SalesforcePlaywright(FakerMixin, BaseLibrary):
    ROBOT_LIBRARY_SCOPE = "Suite"

    def __init__(self):
        super().__init__()
        self._browser = None

    @property
    def browser(self):
        if self._browser is None:
            self._browser = self.builtin.get_library_instance("Browser")
        return self._browser

    def get_current_record_id(self):
        """Parses the current url to get the object id of the current record.
        This expects the url to contain an id that matches [a-zA-Z0-9]{15,18}
        """
        OID_REGEX = r"^(%2F)?([a-zA-Z0-9]{15,18})$"
        url = self.browser.evaluate_javascript(None, "window.location.href")
        for part in url.split("/"):
            oid_match = re.match(OID_REGEX, part)
            if oid_match is not None:
                return oid_match[2]
        raise AssertionError(f"Could not parse record id from url: {url}")

    def go_to_record_home(self, obj_id):
        """Navigates to the Home view of a Salesforce Object

        After navigating, this will wait until the slds-page-header_record-home
        div can be found on the page.
        """
        url = self.cumulusci.org.lightning_base_url
        url = f"{url}/lightning/r/{obj_id}/view"
        self.browser.go_to(url)
        self.wait_until_loading_is_complete("div.slds-page-header_record-home")

    def delete_records_and_close_browser(self):
        """This will close all open browser windows and then delete
        all records that were created with the Salesforce API during
        this testing session.
        """
        self.browser.close_browser("ALL")
        self.salesforce_api.delete_session_records()

    def open_test_browser(
        self, size=None, useralias=None, wait=True, record_video=None
    ):
        """Open a new Playwright browser, context, and page to the default org.

        The return value is a tuple of the browser id, context id, and page details
        returned by the Playwright keywords New Browser, New Context, and New Page.

        This provides the most common environment for testing. For more control,
        you can create your own browser environment with the Browser library
        keywords `Create Browser`, `Create Context`, and `Create Page`.

        To record a video of the session, set `record_video` to True. The video
        (*.webm) will be viewable in the log.html file at the point where this
        keyword is logged.

        This keyword automatically calls the browser keyword `Wait until network is idle`.
        """

        wait = self.builtin.convert_to_boolean(wait)
        default_size = self.builtin.get_variable_value(
            "${DEFAULT BROWSER SIZE}", "1280x1024"
        )
        size = size or default_size

        browser = self.builtin.get_variable_value("${BROWSER}", "chrome")
        headless = browser.startswith("headless")
        browser_type = browser[8:] if headless else browser
        browser_type = "chromium" if browser_type == "chrome" else browser_type
        browser_enum = getattr(SupportedBrowsers, browser_type, None)

        # Note: we can't just pass alias=useralias in the case of useralias being None.
        # That value gets passed to a salesforce query which barfs if the value
        # is None.
        login_url = (
            self.cumulusci.login_url(alias=useralias)
            if useralias
            else self.cumulusci.login_url()
        )

        if record_video:
            # ugh. the "dir" value must be non-empty, and will be treated as
            # a folder name under the browser/video folder. using "../video"
            # seems to be the only way to get the videos to go directly in
            # the video folder. Also, using "." doesn't work :-/
            record_video = {"dir": "../video"}
        width, height = size.split("x", 1)

        browser_id = self.browser.new_browser(browser=browser_enum, headless=headless)
        context_id = self.browser.new_context(
            viewport={"width": width, "height": height}, recordVideo=record_video
        )
        self.browser.set_browser_timeout("15 seconds")
        page_details = self.browser.new_page(login_url)

        if wait:
            self.wait_until_salesforce_is_ready(login_url)
        return browser_id, context_id, page_details

    @capture_screenshot_on_error
    def wait_until_loading_is_complete(self, locator=None, timeout="15 seconds"):
        """Wait for a lightning page to load.

        By default this keyword will wait for any element with the
        class 'slds-template__container', but a different locator can
        be provided.

        In addition to waiting for the element, it will also wait for
        any pending aura events to finish.

        """
        locator = (
            "//div[contains(@class, 'slds-template__container')]/*"
            if locator is None
            else locator
        )
        self.browser.get_elements(locator)

        self.browser.evaluate_javascript(None, WAIT_FOR_AURA_SCRIPT)
        # An old knowledge article recommends waiting a second. I don't
        # like it, but it seems to help. We should do a wait instead,
        # but I can't figure out what to wait on.
        time.sleep(1)

    @capture_screenshot_on_error
    def wait_until_salesforce_is_ready(
        self, login_url, locator=None, timeout="30 seconds"
    ):
        """Attempt to wait until we land on a lightning page

        In addition to waiting for a lightning page, this keyword will
        also attempt to wait until there are no more pending ajax
        requests.

        The timeout parameter is taken as a rough guideline. This
        keyword will actually wait for half of the timeout before
        starting checks for edge cases.

        """

        timeout_seconds = timestr_to_secs(timeout)
        start_time = time.time()

        locator = locator or "div.slds-template__container"
        expected_url = rf"/{self.cumulusci.org.lightning_base_url}\/lightning\/.*/"

        while True:
            try:
                # only wait for half of the timeout before doing some additional
                # checks. This seems to work better than one long timeout.
                self.browser.wait_for_navigation(
                    expected_url, timeout_seconds // 2, PageLoadStates.networkidle
                )
                self.wait_until_loading_is_complete(locator)
                # No errors? We're golden.
                break

            except Exception as exc:
                # dang. Maybe we landed somewhere unexpected?
                if self._check_for_classic():
                    continue

                if time.time() - start_time > timeout_seconds:
                    self.browser.take_screenshot()
                    raise Exception("Timed out waiting for a lightning page") from exc

            # If at first you don't succeed, ...
            self.browser.go_to(login_url)

    def _check_for_classic(self):
        """Switch to lightning if we land on a classic page

        This seems to happen randomly, causing tests to fail
        catastrophically. The idea is to detect such a case and
        auto-click the "switch to lightning" link

        """
        try:
            self.browser.get_element("a.switch-to-lightning")
            self.builtin.log(
                "It appears we are on a classic page; attempting to switch to lightning",
                "WARN",
            )
            # just in case there's a modal present we'll try simulating
            # the escape key. Then, click on the switch-to-lightning link
            self.browser.keyboard_key(KeyAction.press, "Escape")
            self.builtin.sleep("1 second")
            self.browser.click("a.switch-to-lightning")
            return True

        except (AssertionError):
            return False

    def breakpoint(self):
        """Serves as a breakpoint for the robot debugger

        Note: this keyword is a no-op unless the ``robot_debug`` option for
        the task has been set to ``true``. Unless the option has been
        set, this keyword will have no effect on a running test.
        """
        return None



import logging
from pathlib import Path
from unittest import mock

import pytest
import robot.api.logger

from cumulusci.core.config import BaseProjectConfig, UniversalConfig
from cumulusci.core.exceptions import TaskNotFoundError
from cumulusci.core.tests.utils import MockLoggerMixin
from cumulusci.robotframework.CumulusCI import CumulusCI
from cumulusci.tests.util import DummyKeychain, DummyOrgConfig
from cumulusci.utils import temporary_dir


class TestCumulusCILibrary(MockLoggerMixin):
    def setup_method(self):
        self.universal_config = UniversalConfig()
        self.project_config = BaseProjectConfig(
            self.universal_config,
            {
                "project": {"name": "Test"},
                "tasks": {
                    "get_pwd": {
                        "class_path": "cumulusci.tasks.command.Command",
                        "options": {
                            "command": "pwd",
                        },
                    },
                },
                "sources": {
                    "example": {"path": "/tmp"},
                },
            },
            repo_info={"root": Path(__file__).parent.absolute()},
        )
        self.project_config.set_keychain(DummyKeychain())

        self.cumulusci = CumulusCI()
        self.cumulusci._project_config = self.project_config
        self.cumulusci._org = mock.Mock()
        self.cumulusci._org.name = "mock org"
        self.cumulusci._org.instance_url = "https://test.example.com"

        self._task_log_handler.reset()
        self.task_log = self._task_log_handler.messages

    def test_run_task(self):
        """Smoke test; can we run the command task?"""
        result = self.cumulusci.run_task("get_pwd")
        assert result["returncode"] == 0

    def test_run_task_robot_logger(self):
        """Verify that 'run task' uses the robot logger"""
        with mock.patch.object(self.cumulusci, "_run_task"):
            self.cumulusci.run_task("get_pwd")
            args, kwargs = self.cumulusci._run_task.call_args
            task = args[0]
            assert task.logger == robot.api.logger

    def test_robot_logger_supports_warning(self):
        """Verify that 'run task' uses a logger that supports .warning()

        Python deprecated the logger method "warn" in favor of
        "warning". Robot didn't get the memo and has "warn" instead of
        "warning".  Since our tasks use "warning", this verifies that
        we've patched the robot logger before passing it to the task
        constructor.

        """
        with mock.patch.object(self.cumulusci, "_run_task"):
            self.cumulusci.run_task("get_pwd")
            args, kwargs = self.cumulusci._run_task.call_args
            task = args[0]
            assert hasattr(
                task.logger, "warning"
            ), "robot logger should have a warning method but doesn't"

    def test_robot_logger_supports_log(self):
        """Verify that 'run task' uses a logger that supports .log()

        log() normally will be passed a predefined log level (eg:
        logging.INFO, logging.DEBUG, etc), but it can take any integer
        which get mapped to a robot log level as a string. This attempts
        to catch all of the various mappings.
        """
        with mock.patch.object(self.cumulusci, "_run_task"):
            self.cumulusci.run_task("get_pwd")
            args, kwargs = self.cumulusci._run_task.call_args
            task = args[0]
            with mock.patch.object(task.logger, "write") as logger_write:

                task.logger.log(logging.CRITICAL, "a critical message")
                task.logger.log(logging.ERROR, "an error message")
                task.logger.log(logging.WARN, "a warning message")
                task.logger.log(logging.INFO, "an info message")
                task.logger.log(logging.DEBUG, "a debug message")

                task.logger.log(0, "a message with level 0")
                task.logger.log(1, "a message with level 1")
                task.logger.log(11, "a message with level 11")
                task.logger.log(21, "a message with level 21")
                task.logger.log(31, "a message with level 31")
                task.logger.log(41, "a message with level 41")

                logger_write.assert_has_calls(
                    (
                        mock.call("a critical message", "ERROR"),
                        mock.call("an error message", "ERROR"),
                        mock.call("a warning message", "WARN"),
                        mock.call("an info message", "INFO"),
                        mock.call("a debug message", "DEBUG"),
                        mock.call("a message with level 0", "DEBUG"),
                        mock.call("a message with level 1", "DEBUG"),
                        mock.call("a message with level 11", "DEBUG"),
                        mock.call("a message with level 21", "INFO"),
                        mock.call("a message with level 31", "WARN"),
                        mock.call("a message with level 41", "ERROR"),
                    )
                )

    def test_run_task_class_robot_logger(self):
        """Verify that 'run task class' uses the robot logger"""
        with mock.patch.object(self.cumulusci, "_run_task"):
            self.cumulusci.run_task_class(
                "cumulusci.tasks.command.Command", command="ls -l"
            )
            args, kwargs = self.cumulusci._run_task.call_args
            task = args[0]
            assert task.logger == robot.api.logger

    def test_run_unknown_task(self):
        with pytest.raises(TaskNotFoundError):
            self.cumulusci.run_task("bogus")

    def test_cross_project_task(self):
        """Verify that the cross-project task runs with the project config of the task
        See W-8891667
        """
        with temporary_dir() as tmpdir:
            tmpdir = Path(tmpdir).resolve()
            cumulusci_yml_path = tmpdir / "cumulusci.yml"
            with open(cumulusci_yml_path, "w+") as cumulusci_yml:
                self.project_config.sources["example"] = {"path": tmpdir}
                cumulusci_yml.write(
                    """
                    tasks:
                        whatever:
                            class_path: cumulusci.tasks.command.Command
                            options:
                                command: pwd
                    """
                )

            with mock.patch.object(self.cumulusci, "_run_task"):
                self.cumulusci.run_task("example:whatever")

                args, kwargs = self.cumulusci._run_task.call_args
                assert len(args) == 1

                # make sure it's not using the current project config
                # for the task, and that the config it _is_ using is
                # rooted in the directory we created
                task = args[0]
                assert task.project_config != self.cumulusci.project_config
                assert tmpdir == Path(task.project_config.repo_root)

    def test_find_access_token_alias_connected_orgs_multiple_users(self):
        """Verify exception when more than one username matches the criteria"""

        self.cumulusci.org.salesforce_client.query.side_effect = lambda query: {
            "records": [
                {"Username": "tester1@example.com"},
                {"Username": "tester2@example.com"},
            ]
        }
        with pytest.raises(
            Exception,
            match=(
                "More than one user matched the search critiera "
                "for org mock org "
                r"\(tester1@example.com, tester2@example.com\)."
            ),
        ):
            self.cumulusci._find_access_token(
                base_org=self.cumulusci.org, alias="tester1"
            )
            expected_query = "SELECT Username FROM User WHERE alias='tester1'"
            self.cumulusci.org.salesforce_client.query.assert_called_once_with(
                expected_query
            )

    def test_find_access_token_alias_connected_orgs_no_users(self):
        """Verify exception when no username matches the criteria"""

        self.cumulusci.org.salesforce_client.query.side_effect = lambda query: {
            "records": []
        }
        with pytest.raises(
            Exception,
            match=r"Couldn't find a username in org mock org for the specified user \(alias='tester1'\).",
        ):
            self.cumulusci._find_access_token(
                base_org=self.cumulusci.org, alias="tester1"
            )
            expected_query = "SELECT Username FROM User WHERE alias='tester1'"
            self.cumulusci.org.salesforce_client.query.assert_called_once_with(
                expected_query
            )

    def test_find_access_token_alias_connected_orgs_no_matching_org(self):
        self.cumulusci.org.salesforce_client.query.side_effect = lambda query: {
            "records": [
                {"Username": "tester1@example.com"},
            ]
        }

        token = self.cumulusci._find_access_token(
            base_org=self.cumulusci.org, alias="tester1"
        )
        assert token is None

    def test_find_access_token_happy_path(self):
        self.cumulusci.org.salesforce_client.query.side_effect = lambda query: {
            "records": [
                {"Username": "tester1@example.com"},
            ]
        }

        def get_org(org_name):
            configs = {
                "org1": DummyOrgConfig(
                    config={
                        "access_token": "super-secret-token-1",
                        "userinfo": {"preferred_username": "tester1@example.com"},
                    }
                ),
                "org2": DummyOrgConfig(
                    config={
                        "access_token": "super-secret-token-2",
                        "userinfo": {"preferred_username": "tester2@example.com"},
                    }
                ),
            }
            return configs.get(org_name, None)

        def list_orgs():
            return ["org1", "org2"]

        self.cumulusci.keychain.get_org = mock.Mock(wraps=get_org)
        self.cumulusci.keychain.list_orgs = mock.Mock(wraps=list_orgs)

        token = self.cumulusci._find_access_token(
            base_org=self.cumulusci.org, alias="tester1"
        )
        assert token == "super-secret-token-1"

    def test_login_url(self):
        """Verify that login_url by default returns the `start_url` of the org"""
        url = self.cumulusci.login_url()
        assert url == self.cumulusci.org.start_url

    def test_login_url_user_org_with_get_access_token(self):
        """Verify login_url branch when org has a get_access_token method

        This tests a specific branch were we're getting a url for a specific
        user and the org has a `get_access_token` method.
        """
        with mock.patch.object(
            self.cumulusci.org, "get_access_token", return_value="super-secret-token"
        ):

            url = self.cumulusci.login_url(username="test@example.com")
            self.cumulusci.org.get_access_token.assert_called_once_with(
                username="test@example.com"
            )
            assert (
                url
                == "https://test.example.com/secur/frontdoor.jsp?sid=super-secret-token"
            )

    def test_login_url_user_org_without_get_access_token(self):
        """Verify login_url branch when org DOES NOT have a get_access_token method

        This tests a specific branch were we're getting a url for a specific
        user and the org is missing the `get_access_token` method.
        """

        with mock.patch.object(self.cumulusci.org, "get_access_token"):
            self.cumulusci.org.get_access_token = None

            with mock.patch.object(
                self.cumulusci, "_find_access_token", return_value="super-secret-token"
            ):
                self.cumulusci._find_access_token.return_value = "super-secret-token"

                url = self.cumulusci.login_url(username="test@example.com")
                self.cumulusci._find_access_token.assert_called_with(
                    self.cumulusci.org, username="test@example.com"
                )
                assert (
                    url
                    == "https://test.example.com/secur/frontdoor.jsp?sid=super-secret-token"
                )



from .extract_yml import ExtractDeclaration

_DEFAULT_DECLARATIONS = [
    ExtractDeclaration(
        sf_object="Account", where="Name != 'Sample Account for Entitlements'"
    ),
    ExtractDeclaration(sf_object="BusinessHours", where="Name != 'Default'"),
    ExtractDeclaration(sf_object="ContentWorkspace", where="Name != 'Asset Library'"),
    ExtractDeclaration(sf_object="Entitlement", where="Name != 'Sample Entitlement'"),
    ExtractDeclaration(
        sf_object="FieldServiceMobileSettings",
        where="DeveloperName != 'Field_Service_Mobile_Settings'",
    ),
    ExtractDeclaration(
        sf_object="PricebookEntry",
        where="Pricebook2.Id != NULL and Pricebook2.Name != 'Standard Price Book'",
    ),
    ExtractDeclaration(sf_object="Pricebook2", where="Name != 'Standard Price Book'"),
    ExtractDeclaration(
        sf_object="WebLink", where="Name != 'ViewCampaignInfluenceReport'"
    ),
    ExtractDeclaration(
        sf_object="Folder",
        where="DeveloperName NOT IN ('SharedApp', 'EinsteinBotReports')",
    ),
    ExtractDeclaration(
        sf_object="MilestoneType",
        where="Name NOT IN ('First Response to Customer', 'Escalate Case', 'Close Case')",
    ),
    ExtractDeclaration(
        sf_object="WorkBadgeDefinition",
        where="Name NOT IN ('Thanks', 'You\\'re a RockStar!', 'Team Player',"
        "'All About Execution', 'Teacher', 'Top Performer', 'Hot Lead',"
        "'Key Win', 'Customer Hero', 'Competition Crusher',"
        "'Deal Maker', 'Gold Star')",
    ),
    ExtractDeclaration(
        sf_object="EmailTemplate",
        where="DeveloperName NOT IN ('CommunityLockoutEmailTemplate',"
        "'CommunityVerificationEmailTemplate',"
        "'CommunityChgEmailVerOldTemplate',"
        "'CommunityChgEmailVerNewTemplate',"
        "'CommunityDeviceActEmailTemplate',"
        "'CommunityWelcomeEmailTemplate',"
        "'CommunityChangePasswordEmailTemplate',"
        "'CommunityForgotPasswordEmailTemplate' )",
    ),
]
DEFAULT_DECLARATIONS = {decl.sf_object: decl for decl in _DEFAULT_DECLARATIONS}



import os
import webbrowser
from functools import lru_cache
from typing import List, Optional, Union
from urllib.parse import urlparse

import github3
from github3 import GitHub, GitHubEnterprise  # , login
from github3.exceptions import (
    AuthenticationFailed,
    ConnectionError,
    ForbiddenError,
    NotFoundError,
    ResponseError,
    TransportError,
)
from github3.session import GitHubSession
from requests.adapters import HTTPAdapter
from requests.exceptions import RetryError
from requests.models import Response
from requests.packages.urllib3.util.retry import Retry

from cumulusci.core.config import BaseProjectConfig, ServiceConfig
from cumulusci.core.exceptions import (  # DependencyLookupError
    GithubApiError,
    GithubApiNotFoundError,
    GithubException,
    ServiceNotConfigured,
)
from cumulusci.tasks.github.util import CommitDir
from cumulusci.utils.git import parse_repo_url
from cumulusci.vcs.base import VCSService
from cumulusci.vcs.github import GitHubRelease, GitHubRepository
from cumulusci.vcs.github.release_notes.generator import (
    GithubReleaseNotesGenerator,
    ParentPullRequestNotesGenerator,
)
from cumulusci.vcs.github.release_notes.parser import parser_configs

OAUTH_DEVICE_APP = {
    "client_id": "2a4bc3e5ce4f2c49a957",
    "auth_uri": "https://github.com/login/device/code",
    "token_uri": "https://github.com/login/oauth/access_token",
    "scope": "repo gist",
}
SSO_WARNING = """Results may be incomplete. You have not granted your Personal Access token access to the following organizations:"""
UNAUTHORIZED_WARNING = """
Bad credentials. Verify that your personal access token is correct and that you are authorized to access this resource.
"""
SELF_SIGNED_WARNING = """
There was a problem verifying the SSL Certificate due to a certificate authority that isn't trusted or a self-signed certificate in the certificate chain. Try setting CUMULUSCI_SYSTEM_CERTS Environment Variable to 'True'. See https://cumulusci.readthedocs.io/en/stable/env-var-reference.html?#cumulusci-system-certs
"""


class GitHubRety(Retry):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def increment(self, *args, **kwargs):
        # Check for connnection and fail on SSLerror
        # SSLCertVerificationError
        if "error" in kwargs:
            error = kwargs["error"]
            error_str = "CERTIFICATE_VERIFY_FAILED"
            if error_str in str(error):
                raise error
        # finally call increment
        return super().increment(*args, **kwargs)


# Prepare request retry policy to be attached to github sessions.
# 401 is a weird status code to retry, but sometimes it happens spuriously
# and https://github.community/t5/GitHub-API-Development-and/Random-401-errors-after-using-freshly-generated-installation/m-p/22905 suggests retrying
retries = GitHubRety(status_forcelist=(401, 502, 503, 504), backoff_factor=0.3)
adapter = HTTPAdapter(max_retries=retries)

INSTALLATIONS = {}


def _determine_github_client(host: Union[str, None], client_params: dict) -> GitHub:
    """Determine the appropriate GitHub client based on the host.

    Args:
        host (Union[str, None]): The host for the GitHub client.
        client_params (dict): Parameters for the GitHub client.

    Returns:
        GitHub: The GitHub client instance.
    """
    # also covers "api.github.com"
    is_github: bool = host in (None, "None") or "github.com" in host
    client_cls: GitHub = GitHub if is_github else GitHubEnterprise  # type: ignore
    params: dict = client_params
    if not is_github:
        params["url"] = "https://" + host  # type: ignore

    return client_cls(**params)


def get_github_api_for_repo(keychain, repo_url, session=None) -> GitHub:
    owner, repo_name, host = parse_repo_url(repo_url)
    gh: GitHub = _determine_github_client(
        host,
        {
            "session": session
            or GitHubSession(default_read_timeout=30, default_connect_timeout=30)
        },
    )

    # Apply retry policy
    gh.session.mount("http://", adapter)
    gh.session.mount("https://", adapter)

    GITHUB_TOKEN = os.environ.get("GITHUB_TOKEN")
    APP_KEY = os.environ.get("GITHUB_APP_KEY", "").encode("utf-8")
    APP_ID = os.environ.get("GITHUB_APP_ID")
    if APP_ID and APP_KEY:
        installation = INSTALLATIONS.get((owner, repo_name))
        if installation is None:
            gh.login_as_app(APP_KEY, APP_ID, expire_in=120)
            try:
                installation = gh.app_installation_for_repository(owner, repo_name)
            except github3.exceptions.NotFoundError:
                raise GithubException(
                    f"Could not access {owner}/{repo_name} using GitHub app. "
                    "Does the app need to be installed for this repository?"
                )
            INSTALLATIONS[(owner, repo_name)] = installation
        gh.login_as_app_installation(APP_KEY, APP_ID, installation.id)
    elif GITHUB_TOKEN:
        gh.login(token=GITHUB_TOKEN)
    else:
        token = get_auth_from_service(host, keychain)
        gh.login(token=token)

    return gh


def get_auth_from_service(host, keychain) -> str:
    """
    Given a host extracted from a repo_url, returns the username and token for
    the first service with a matching server_domain
    """
    if host is None or host == "None" or "github.com" in host:
        service_config = keychain.get_service("github")
    else:
        services = keychain.get_services_for_type("github_enterprise")
        service_by_host = {service.server_domain: service for service in services}

        # Check when connecting to server, but not when creating new service as this would always catch
        if list(service_by_host.keys()).count(host) == 0:
            raise ServiceNotConfigured(
                f"No Github Enterprise service configured for domain {host}."
            )

        service_config = service_by_host[host]

    # Basic Auth no longer supported on github.com, so only returning token
    # this requires GitHub Enterprise to use token auth and not Basic auth
    # docs.github.com/en/rest/overview/other-authentication-methods#via-username-and-password
    return service_config.token


def validate_gh_enterprise(host: str, keychain) -> None:
    services = keychain.get_services_for_type("github_enterprise")
    if services:
        hosts = [service.server_domain for service in services]
        if hosts.count(host) > 1:
            raise GithubException(
                f"More than one Github Enterprise service configured for domain {host}."
            )


def get_sso_disabled_orgs(response: Response) -> list:
    """
    Given a response from Github, return a list of organization IDs without SSO
    grants.
    """
    disabled_orgs = []
    sso_header = response.headers.get("X-Github-Sso")
    partial_results_prefix = "partial-results; organizations="

    if sso_header and partial_results_prefix in sso_header:
        disabled_orgs = sso_header[len(partial_results_prefix) :].split(",")

    return disabled_orgs


def format_github3_exception(
    exc: Union[ResponseError, TransportError, ConnectionError]
) -> str:
    """Checks github3 exceptions for the most common GitHub authentication
    issues, returning a user-friendly message if found.

    @param exc: The exception to process
    @returns: The formatted exception string
    """
    user_warning = ""

    too_many_str = "too many 401 error responses"
    is_bad_auth_retry = (
        type(exc) is TransportError
        and type(exc.exception) is RetryError
        and too_many_str in str(exc.exception)
    )
    is_auth_failure = type(exc) is AuthenticationFailed

    if is_bad_auth_retry or is_auth_failure:
        user_warning = UNAUTHORIZED_WARNING

    if isinstance(exc, ResponseError):
        scope_error_msg = check_github_scopes(exc)
        sso_error_msg = check_github_sso_auth(exc)
        user_warning = scope_error_msg + sso_error_msg

    if isinstance(exc, ConnectionError):
        if "self signed certificate" in str(exc.exception):
            user_warning = SELF_SIGNED_WARNING
        else:
            return ""

    return user_warning


def get_oauth_scopes(response: Response) -> set:
    """
    Given a response from Github, return the set of OAuth scopes for its
    request.
    """
    authorized_scopes = set()

    # If the token isn't authorized "X-OAuth-Scopes" header won't be present
    x_oauth_scopes = response.headers.get("X-OAuth-Scopes")
    if x_oauth_scopes:
        authorized_scopes = set(x_oauth_scopes.split(", "))

    return authorized_scopes


def check_github_scopes(exc: ResponseError) -> str:
    """
    Parse github3 ResponseError headers for the correct scopes and return a
    warning if the user is missing.

    @param exc: The exception to process
    @returns: The formatted exception string
    """

    user_warning = ""

    has_wrong_status_code = exc.response.status_code not in (403, 404)
    if has_wrong_status_code:
        return user_warning

    token_scopes = get_oauth_scopes(exc.response)

    # Gist resource won't return X-Accepted-OAuth-Scopes for some reason, so this
    # string might be `None`; we discard the empty string if so.
    accepted_scopes = exc.response.headers.get("X-Accepted-OAuth-Scopes") or ""
    accepted_scopes = set(accepted_scopes.split(", "))
    accepted_scopes.discard("")

    request_url = urlparse(exc.response.url)
    if not accepted_scopes and request_url.path == "/gists":
        accepted_scopes = {"gist"}

    missing_scopes = accepted_scopes.difference(token_scopes)
    if missing_scopes:
        user_warning = f"Your token may be missing the following scopes: {', '.join(missing_scopes)}\n"
        # This assumes we're not on enterprise and 'api.github.com' == request_url.hostname
        user_warning += (
            "Visit Settings > Developer settings > Personal access tokens to add them."
        )

    return user_warning


def check_github_sso_auth(exc: ResponseError) -> str:
    """
    Check ResponseError header for SSO authorization and return a warning if
    required

    @param exc: The exception to process
    @returns: The formatted exception string
    """
    user_warning = ""
    headers = exc.response.headers

    if exc.response.status_code != 403 or "X-Github-Sso" not in headers:
        return user_warning

    sso_header = str(headers["X-Github-Sso"] or "")
    if sso_header.startswith("required; url="):
        # In this case the message from github is good enough, but we can help
        # the user by opening a browser to authorize the token.
        auth_url = sso_header.split("url=", maxsplit=1)[1]
        user_warning = f"{exc.message}\n{auth_url}"
        webbrowser.open(auth_url)
    elif sso_header.startswith("partial-results"):
        # In cases where we don't have complete results we get the
        # partal-results header, so return the organization IDs. This may or
        # may not be useful without help from us to lookup the org IDs.
        unauthorized_org_ids = get_sso_disabled_orgs(exc.response)
        user_warning = f"{SSO_WARNING} {unauthorized_org_ids}"

    return user_warning


@lru_cache(50)
def get_github_service_for_url(
    project_config: BaseProjectConfig, url: str
) -> Optional[VCSService]:
    return GitHubService.get_service_for_url(
        project_config, url
    ) or GitHubEnterpriseService.get_service_for_url(project_config, url)


class GitHubService(VCSService):
    service_type = "github"
    _repo: GitHubRepository
    github: GitHub

    def __init__(self, config: BaseProjectConfig, name: Optional[str] = None, **kwargs):
        """Initializes the GitHub service with the given project configuration.
        Args:
            config (BaseProjectConfig): The configuration for the GitHub service.
            name (str): The name or alias of the VCS service.
            **kwargs: Additional keyword arguments.
        """
        super().__init__(config, name=name, **kwargs)
        repo_url = kwargs.get("repository_url", self.config.repo_url)
        self.github = get_github_api_for_repo(self.keychain, repo_url)
        self._repo: GitHubRepository = None

    @property
    def dynamic_dependency_class(self):  # -> Type[GitHubDynamicDependency]:
        """Returns the dynamic dependency class for the GitHub service."""
        from cumulusci.core.dependencies.github import GitHubDynamicDependency

        return GitHubDynamicDependency

    @property
    def repo(self) -> GitHubRepository:
        """Returns the GitHub repository associated with the service."""
        return self._repo

    @repo.setter
    def repo(self, repo: GitHubRepository):
        """Set the GitHub repository associated with the service.

        Args:
            repo (GitHubRepository): The GitHub repository instance to set.
        """
        self._repo = repo

    @classmethod
    def validate_service(cls, options: dict, keychain) -> dict:
        """Validates service for Github and GithubEnterprise.

        Args:
            options (dict): The options for the service validation.
            keychain: The keychain for accessing project credentials.

        Returns:
            dict: The validated options for the service.
        """
        username = options["username"]
        token = options["token"]
        # Github service doesn't have "server_domain",
        server_domain = options.get("server_domain", None)

        gh = _determine_github_client(server_domain, {"token": token})
        if isinstance(gh, GitHubEnterprise):
            validate_gh_enterprise(server_domain, keychain)
        try:
            authed_user = gh.me()
            auth_login = authed_user.login
            assert username == auth_login, f"{username}, {auth_login}"
        except AssertionError as e:
            raise GithubException(
                f"Service username and token username do not match. ({str(e)})"
            )
        except Exception as e:
            warning_msg = format_github3_exception(e) or str(e)
            raise GithubException(
                f"Could not confirm access to the GitHub API: {warning_msg}"
            )
        else:
            member_orgs = {f"{org.id}": f"{org.login}" for org in gh.organizations()}
            options["Organizations"] = ", ".join([k for k in member_orgs.values()])

            # We're checking for a partial-response SSO header and /user/orgs
            # doesn't include one, so we need /user/repos instead.
            repo_generator = gh.repositories()
            _ = next(repo_generator, None)
            repo_response = repo_generator.last_response
            options["scopes"] = ", ".join(sorted(get_oauth_scopes(repo_response)))

            unauthorized_org_ids = get_sso_disabled_orgs(repo_response)
            unauthorized_orgs = {
                k: member_orgs[k] for k in unauthorized_org_ids if k in member_orgs
            }
            if unauthorized_orgs:
                options["SSO Disabled"] = ", ".join(
                    [k for k in unauthorized_orgs.values()]
                )

            expiration_date = repo_response.headers.get(
                "GitHub-Authentication-Token-Expiration"
            )
            if expiration_date:
                options["expires"] = expiration_date

        return options

    @classmethod
    def get_service_for_url(
        cls,
        project_config: BaseProjectConfig,
        url: str,
        service_alias: Optional[str] = None,
    ) -> Optional["GitHubService"]:
        """Returns the service configuration for the given URL."""
        _owner, _repo_name, host = parse_repo_url(url)

        if host is None or host == "None" or "github.com" in host:
            service_config = project_config.keychain.get_service(
                cls.service_type, service_alias
            )

            vcs_service = GitHubService(
                project_config,
                name=service_config.name,
                service_config=service_config,
                logger=project_config.logger,
                repository_url=url,
            )
            project_config.logger.info(
                f"Github service configured for domain {host} : {url}."
            )
            return vcs_service
        project_config.logger.debug(
            f"No Github service configured for domain {host} : {url}."
        )
        return None

    def get_repository(self, options: dict = {}) -> GitHubRepository:
        """Returns the GitHub repository."""
        try:
            if self._repo is None:
                self._repo = GitHubRepository(
                    self.github,
                    self.config,
                    logger=self.logger,
                    service_type=self.service_type,
                    service_config=self.service_config,
                    options=options,
                )
        except NotFoundError as e:
            raise GithubApiNotFoundError(f"GitHub repository not found: {e}")
        except ForbiddenError as e:
            raise GithubApiError(f"GitHub repository is not accessible: {e}")
        return self.repo

    def parse_repo_url(self) -> List[str]:
        owner, repo_name, host = parse_repo_url(self.repo_url)
        return [host or "", owner or "", repo_name or ""]

    def get_committer(self, repo: GitHubRepository) -> CommitDir:
        """Returns the committer for the GitHub repository."""
        return CommitDir(repo.repo, logger=self.logger)

    def markdown(
        self, release: GitHubRelease, mode: str = "gfm", context: str = ""
    ) -> str:
        """Converts the given text to GitHub-flavored Markdown."""
        release_html = self.github.markdown(
            release,
            mode=mode,
            context=context,
        )
        return release_html

    def release_notes_generator(self, options: dict) -> GithubReleaseNotesGenerator:
        github_info = {
            "github_owner": self.config.repo_owner,
            "github_repo": self.config.repo_name,
            "github_username": self.service_config.username,
            "github_password": self.service_config.password,
            "default_branch": self.config.project__git__default_branch,
            "prefix_beta": self.config.project__git__prefix_beta,
            "prefix_prod": self.config.project__git__prefix_release,
        }

        generator = GithubReleaseNotesGenerator(
            self.github,
            github_info,
            parser_configs(self.config),
            options["tag"],
            options.get("last_tag"),
            options.get("link_pr"),
            options.get("publish"),
            self.get_repository().has_issues,
            options.get("include_empty"),
            version_id=options.get("version_id"),
            trial_info=options.get("trial_info", False),
            sandbox_date=options.get("sandbox_date", None),
            production_date=options.get("production_date", None),
        )

        return generator

    def parent_pr_notes_generator(
        self, repo: GitHubRepository
    ) -> ParentPullRequestNotesGenerator:
        """Returns the parent pull request notes generator for the GitHub repository."""
        return ParentPullRequestNotesGenerator(self.github, repo.repo, self.config)


class GitHubEnterpriseService(GitHubService):
    service_type = "github_enterprise"
    _repo: GitHubRepository

    def __init__(self, config: BaseProjectConfig, name: Optional[str] = None, **kwargs):
        super().__init__(config, name=name, **kwargs)

    @property
    def dynamic_dependency_class(self):  # -> Type[GitHubDynamicDependency]:
        """Returns the dynamic dependency class for the GitHub service."""
        from cumulusci.core.dependencies.github import GitHubDynamicDependency

        return GitHubDynamicDependency

    @classmethod
    def get_service_for_url(
        cls,
        project_config: BaseProjectConfig,
        url: str,
        service_alias: Optional[str] = None,
    ) -> Optional["GitHubEnterpriseService"]:
        """Returns the service configuration for the given URL."""
        _owner, _repo_name, host = parse_repo_url(url)

        configured_services: list[
            ServiceConfig
        ] = project_config.keychain.get_services_for_type(cls.service_type)
        service_by_host = {
            service.server_domain: service for service in configured_services
        }

        # Check when connecting to server, but not when creating new service as this would always catch
        if list(service_by_host.keys()).count(host) == 0:
            project_config.logger.debug(
                f"No Github Enterprise service configured for domain {host} : {url}."
            )
            return None

        service_config = service_by_host[host]
        vcs_service = GitHubEnterpriseService(
            project_config,
            name=service_config.name,
            service_config=service_config,
            logger=project_config.logger,
            repository_url=url,
        )
        project_config.logger.info(
            f"Github Enterprise service configured for domain {host} : {url}."
        )
        return vcs_service



import inspect
import logging
import os
from abc import ABC
from pathlib import Path
from typing import Optional

from cumulusci.utils.yaml.cumulusci_yml import Plugin, cci_safe_load


class PluginBase(ABC):
    plugin: Plugin
    plugin_config_file: str = "cumulusci_plugin.yml"
    path: str
    plugin_project_config: Optional[dict] = None

    def __init__(self, **kwargs) -> None:
        self.logger = kwargs.get("logger", logging.getLogger(__name__))
        self.path = inspect.getfile(self.__class__)
        self._load_config()

    @property
    def api_name(self) -> str:
        """Returns the api_name of the plugin."""
        return self._api_name

    @api_name.setter
    def api_name(self, value: str) -> None:
        """Sets the api_name of the plugin."""
        self._api_name = value

    @property
    def name(self) -> str:
        """Returns the name of the plugin."""
        return (
            self.plugin.name if self.plugin and self.plugin.name else "Unnamed Plugin"
        )

    @property
    def version(self) -> str:
        """Returns the version of the plugin."""
        return self.plugin.version if self.plugin and self.plugin.version else "0.0.0"

    def initialize(self) -> None:
        """Initialize the plugin. This method is called when the plugin is loaded."""
        self.logger.info(
            f"Initializing plugin: {self.name}) v{self.version} by {self.plugin.author}"
        )
        self._initialize()
        self.logger.info(f"Loaded plugin: {self.name}) v{self.version}")

    def teardown(self) -> None:
        """Tear down the plugin. This method is called when the plugin is unloaded."""
        self.logger.info(f"Tearing down plugin: {self.name})")

    @property
    def config_plugin_path(self) -> Optional[str]:
        """Returns the path to the plugin configuration file."""
        plugin_config_path = os.path.join(
            os.path.dirname(self.path), self.plugin_config_file
        )
        plugin_path = Path(plugin_config_path)
        if plugin_path.is_file():
            return str(plugin_path)
        return None

    def _load_config(self) -> None:
        """Loads the plugin configuration file if it exists."""
        config_plugin_path = self.config_plugin_path
        if config_plugin_path:
            self.logger.info(
                f"{self.__class__.__name__}: Plugin configuration file found at: {config_plugin_path}"
            )
            self.plugin_project_config = cci_safe_load(
                config_plugin_path, context=self.__class__.__name__, logger=self.logger
            )

            for plugin_name in self.plugin_project_config.get("plugins", {}).keys():
                self._api_name = plugin_name
                self.plugin = Plugin(
                    **self.plugin_project_config.get("plugins", {}).get(plugin_name, {})
                )
                break

            self.logger.info(
                f"{self.name}: Plugin project configuration loaded successfully."
            )
        else:
            self.logger.warning(
                f"{self.__class__.__name__}: No plugin configuration file found."
            )

    def _initialize(self) -> None:
        """Override this method to add custom initialization logic."""
        pass

    def check_latest_version(self):
        """Override this method to check for the latest version of the plugin."""
        # Implement logic to check for the latest version if needed
        pass



from typing import List, Optional, Set
from urllib.parse import quote

from pydantic.v1 import BaseModel

from cumulusci.tasks.metadata_etl import MetadataSingleEntityTransformTask
from cumulusci.utils import inject_namespace
from cumulusci.utils.xml.metadata_tree import MetadataElement


def _inject_namespace(text: str, options: dict) -> str:
    return inject_namespace(
        "",
        text,
        namespace=options["namespace_inject"]
        if not options.get("namespaced_org")
        else "",
        managed=options.get("managed") or False,
        namespaced_org=options.get("namespaced_org"),
    )[1]


class ProfileActionOverrideOptions(BaseModel):
    """Options for a single profileActionOverride"""

    action_name: str
    content: str
    form_factor: str
    page_or_sobject_type: str
    record_type: Optional[str]
    type: str
    profile: str
    options: dict = {}

    def namespace_inject(self, field_name: str) -> None:
        setattr(
            self, field_name, _inject_namespace(getattr(self, field_name), self.options)
        )


class AddProfileActionOverridesOptions(BaseModel):
    """Options container for profile action overrides"""

    name: str
    overrides: List[ProfileActionOverrideOptions]


class AddProfileActionOverrides(MetadataSingleEntityTransformTask):
    """
    Inserts or updates profileActionOverrides in CustomApplication metadata.
    If a profileActionOverride with the same actionName, pageOrSobjectType,
    recordType, and profile already exists, it will be updated with a warning.
    Otherwise, a new override will be added.
    Task option details:
    - applications: List of CustomApplications to modify
        - name: API name of the CustomApplication to modify
        - overrides: List of profile action overrides to add/update
            - action_name: Action name (e.g., "View", "Edit", "New")
            - content: Content reference (page/component API name)
            - form_factor: Form factor (e.g., "Large", "Small")
            - page_or_sobject_type: Page or SObject type
            - record_type: Record type API name (optional)
            - type: Override type (e.g., "Flexipage", "Visualforce", "LightningComponent")
            - profile: Profile name or API name
    Example Usage
    -----------------------
    .. code-block::  yaml
        task: add_profile_action_overrides
        options:
            applications:
                - name: "%%%NAMESPACE%%%CustomApplicationConsole"
                  overrides:
                    - action_name: View
                      content: "%%%NAMESPACED_ORG%%%AccountUserRecordPage"
                      form_factor: Large
                      page_or_sobject_type: Account
                      record_type: PersonAccount.User
                      type: Flexipage
                      profile: Admin
    """

    entity = "CustomApplication"
    task_options = {
        "applications": {
            "description": "List of CustomApplications to modify. See task info for structure.",
            "required": True,
        },
        **MetadataSingleEntityTransformTask.task_options,
    }

    def _init_options(self, kwargs):
        super()._init_options(kwargs)

        # Validate options using Pydantic
        self._validated_options: List[AddProfileActionOverridesOptions] = []
        for application in self.options.get("applications"):
            validated_options = AddProfileActionOverridesOptions(
                name=quote(_inject_namespace(application.get("name"), self.options)),
                overrides=application.get("overrides"),
            )

            self._validated_options.append(validated_options)

        self.api_names: Set[str] = set(
            application.name for application in self._validated_options
        )

    def _transform_entity(
        self, metadata: MetadataElement, api_name: str
    ) -> Optional[MetadataElement]:

        if not self._validated_options:
            self.logger.warning("No applications to add profile action overrides for")
            return None

        for application in self._validated_options:
            if application.name != api_name:
                continue

            for override_config in application.overrides:
                self._add_or_update_override(
                    metadata, application.name, override_config
                )

        return metadata

    def _add_or_update_override(self, metadata, api_name, override_config):
        """Add or update a single profileActionOverride"""
        override_config.options = self.options

        # Inject namespace where needed
        override_config.namespace_inject("content")
        override_config.namespace_inject("page_or_sobject_type")
        override_config.namespace_inject(
            "record_type"
        ) if override_config.record_type else None

        # Find existing override with same key fields
        existing_override = self._find_existing_override(metadata, override_config)

        if existing_override:
            self.logger.warning(
                f"Updating existing profileActionOverride for {override_config.profile}/{override_config.page_or_sobject_type}/{override_config.action_name} in {api_name}"
            )
            # Update the existing override
            self._update_override_element(
                existing_override,
                override_config,
            )
        else:
            self.logger.info(
                f"Adding profileActionOverride for {override_config.profile}/{override_config.page_or_sobject_type}/{override_config.action_name} to {api_name}"
            )
            # Create new override
            self._create_new_override(metadata, override_config)

    def _find_existing_override(self, metadata, override_config):
        """
        Find an existing profileActionOverride that matches the key fields:
        actionName, pageOrSobjectType, recordType, and profile
        """
        for override_elem in metadata.findall("profileActionOverrides"):
            elem_action = override_elem.find("actionName")
            elem_page = override_elem.find("pageOrSobjectType")
            elem_record_type = override_elem.find("recordType")
            elem_profile = override_elem.find("profile")

            # Match on all key fields
            if (
                elem_action
                and elem_action.text == override_config.action_name
                and elem_page
                and elem_page.text == override_config.page_or_sobject_type
                and elem_profile
                and elem_profile.text == override_config.profile
            ):
                # Handle recordType - both must be None or both must match
                if override_config.record_type is None and elem_record_type is None:
                    return override_elem
                elif (
                    override_config.record_type is not None
                    and elem_record_type is not None
                    and elem_record_type.text == override_config.record_type
                ):
                    return override_elem

        return None

    def _update_override_element(
        self,
        override_elem,
        override_config,
    ):
        """Update an existing profileActionOverride element"""
        # Update each child element
        # actionName
        elem = override_elem.find("actionName")
        if elem is not None:
            elem.text = override_config.action_name

        # content
        elem = override_elem.find("content")
        if elem is not None:
            elem.text = override_config.content

        # formFactor
        elem = override_elem.find("formFactor")
        if elem is not None:
            elem.text = override_config.form_factor

        # pageOrSobjectType
        elem = override_elem.find("pageOrSobjectType")
        if elem is not None:
            elem.text = override_config.page_or_sobject_type

        # recordType (optional)
        elem = override_elem.find("recordType")
        if override_config.record_type:
            if elem is not None:
                elem.text = override_config.record_type
        elif elem is not None:
            # Remove recordType if it exists but shouldn't
            override_elem.remove(elem)

        # type
        elem = override_elem.find("type")
        if elem is not None:
            elem.text = override_config.type

        # profile
        elem = override_elem.find("profile")
        if elem is not None:
            elem.text = override_config.profile

    def _create_new_override(
        self,
        metadata,
        override_config,
    ):
        """Create a new profileActionOverride element with proper ordering"""
        override_elem = metadata.append("profileActionOverrides")

        # Add elements in the correct order per Salesforce metadata API
        override_elem.append("actionName", text=override_config.action_name)
        override_elem.append("content", text=override_config.content)
        override_elem.append("formFactor", text=override_config.form_factor)
        override_elem.append(
            "pageOrSobjectType", text=override_config.page_or_sobject_type
        )

        # recordType is optional
        if override_config.record_type:
            override_elem.append("recordType", text=override_config.record_type)

        override_elem.append("type", text=override_config.type)
        override_elem.append("profile", text=override_config.profile)



import os
import subprocess
import sys
import typing as T
from json import JSONDecodeError

from cumulusci.core.exceptions import CumulusCIException


def safe_json_from_response(response):
    "Check JSON response is HTTP200 and actually JSON."
    response.raise_for_status()

    try:
        return response.json()
    except JSONDecodeError:
        raise CumulusCIException(f"Cannot decode as JSON:  {response.text}")


is_trust_patched = False


def init_requests_trust():
    """Monkey-patch requests to ensure we validate certificates using the preferred trust store.

    Currently this does nothing unless the CUMULUSCI_SYSTEM_CERTS environment variable is set to True.
    (That may change so that this is done by default, after further testing.)

    This is called from cumulusci.__init__ to enact our policy:

    - On all platforms, block `requests` from passing the `certifi` CA bundle
      to urllib3, so that urllib3 will instead set up an SSLContext by
      running SSLContext.load_default_certs()

    - On macOS, override SSLContext.load_default_certs() to load certs
      from the system keychain instead of from OpenSSL's default path.

    If a request explicitly sets `verify` to a path
    (perhaps via the REQUESTS_CA_BUNDLE environment variable),
    the CA certs from that path will still be trusted as well.

    The monkey-patching approach may be controversial, but it ensures that:
    a. we don't have to change every location we are using requests.get
       or requests.post without an explicit session
    b. our policy will also apply to 3rd-party libraries that use requests
    """
    if os.environ.get("CUMULUSCI_SYSTEM_CERTS") != "True":
        return
    global is_trust_patched
    if is_trust_patched:
        return
    is_trust_patched = True

    from requests.adapters import HTTPAdapter
    from requests.utils import DEFAULT_CA_BUNDLE_PATH

    # On macOS, monkey patch SSLContext.load_default_locations
    # to load CA certs from the system keychain
    if sys.platform == "darwin":
        import ssl

        cadata = get_macos_ca_certs()

        def load_default_certs(self, purpose=ssl.Purpose.SERVER_AUTH):
            self.load_verify_locations(cadata=cadata)

        ssl.SSLContext.load_default_certs = load_default_certs

    # On all platforms, monkey patch HTTPAdapter.cert_verify
    #  to avoid using the CA bundle from certifi

    orig_cert_verify = HTTPAdapter.cert_verify

    def cert_verify(self, conn, *args, **kw):
        orig_cert_verify(self, conn, *args, **kw)
        if conn.ca_certs == DEFAULT_CA_BUNDLE_PATH:
            conn.ca_certs = None

    HTTPAdapter.cert_verify = cert_verify


def get_certs_from_keychain(path: T.Optional[str] = None) -> str:
    """Get certs from the specified macOS keychain path (or default keychains if path is None)"""
    args = [
        "security",
        "find-certificate",
        "-a",
        "-p",
    ]
    if path:
        args.append(path)
    return subprocess.run(
        args,
        stdout=subprocess.PIPE,
        encoding="latin-1",
    ).stdout


def get_macos_ca_certs() -> str:
    """Get certs from both the default keychains and the SystemRootCertificates keychain"""
    return get_certs_from_keychain(None) + get_certs_from_keychain(
        "/System/Library/Keychains/SystemRootCertificates.keychain"
    )



# discovered by trial and error. Usually extended with a list of
# tooling objects. i.e. any object which is both a tooling object
# and also a "regular" sobject will be skipped.
OPT_IN_ONLY = [
    "ApexClass",
    "ApexTrigger",
    "FeedItem",
    "Translation",
    "WebLinkLocalization",
    "RecordTypeLocalization",
    "RecordType",
    "BrandTemplate",
]

NOT_COUNTABLE = (
    "ContentDocumentLink",  # ContentDocumentLink requires a filter by a single Id on ContentDocumentId or LinkedEntityId
    "ContentFolderItem",  # Implementation restriction: ContentFolderItem requires a filter by Id or ParentContentFolderId
    "ContentFolder",  # Similar to above
    "ContentFolderLink",  # Similar to above
    "ContentFolderMember",  # Similar to above
    "IdeaComment",  # you must filter using the following syntax: CommunityId = [single ID],
    "Vote",  # you must filter using the following syntax: ParentId = [single ID],
    "RecordActionHistory",  # Gack: 1133111327-118855 (1126216936)
    "DashboardSnapshotResults",  # Gack: 1198622208-46932 (1126216936)
    "RecordRecommendation",  # Implementation restriction: RecordRecommendation requires a filter on TargetSobjectType or TargetId or RecordRecommendationId
)


NOT_EXTRACTABLE = NOT_COUNTABLE + (
    "%Share",
    "%Access",
    "%History",
    "%Permission",
    "%PermissionSet",
    "%Permissions",
    "AuthorizationFormDataUse",
    "CustomHelpMenuSection",
    "DataUseLegalBasis",
    "DataUsePurpose",
    "ExternalDataUserAuth",
    "FieldPermissions",
    "Group",
    "GroupMember",
    "PermissionSet",
    "PermissionSetAssignment",
    "PermissionSetGroup",
    "PermissionSetGroupComponent",
    "PermissionSetLicenseAssign",
    "PermissionSetTabSetting",
    "Profile",
    "RecordType",
    "User",
    "UserAppInfo",
    "UserAppMenuCustomization",
    "UserCustomBadge",
    "UserCustomBadgeLocalization",
    "UserEmailPreferredPerson",
    "UserListView",
    "UserListViewCriterion",
    "UserPackageLicense",
    "UserPreference",
    "UserProvAccount",
    "UserProvAccountStaging",
    "UserProvMockTarget",
    "UserProvisioningConfig",
    "UserProvisioningLog",
    "UserProvisioningRequest",
    "UserRole",
)

# Generated with these patterns:
#     "%permission%",
#     "%use%",
#     "%access%",
#     "group",
#     "%share",
#     "NetworkUserHistoryRecent",
#     "ObjectPermissions",
#     "OmniSupervisorConfigUser",
#     "OutgoingEmail",
#     "OutgoingEmailRelation",
# )

# And this code:
#
# patterns_to_ignore = NOT_EXTRACTABLE_WITHOUT_NOT_COUNTABLE

# names = [obj["name"] for obj in sf.describe()["sobjects"] if obj["createable"]]

# regexps_to_ignore = [
#     re.compile(pat.replace("%", ".*"), re.IGNORECASE) for pat in patterns_to_ignore
# ]

# for name in names:
#     if any(reg.match(name) for reg in regexps_to_ignore) and not ("__" in name):
#         print(name)



"""Tests for directoryRecreator module."""

import os
import tempfile
from pathlib import Path
from unittest import mock

import pytest

from cumulusci.core.config import TaskConfig
from cumulusci.core.exceptions import TaskOptionsError
from cumulusci.tasks.utility.directoryRecreator import DirectoryRecreator


class TestDirectoryRecreatorOptions:
    """Test cases for DirectoryRecreator options configuration."""

    def test_required_path_option(self):
        """Test that path option is required."""
        with tempfile.TemporaryDirectory() as tmpdir:
            test_dir = os.path.join(tmpdir, "test_directory")
            task_config = TaskConfig({"options": {"path": test_dir}})

            task = DirectoryRecreator(
                project_config=mock.Mock(),
                task_config=task_config,
                org_config=None,
            )

            assert task.parsed_options.path == Path(test_dir)

    def test_path_as_string(self):
        """Test initialization with path as string."""
        with tempfile.TemporaryDirectory() as tmpdir:
            test_dir = os.path.join(tmpdir, "test_dir")
            task_config = TaskConfig({"options": {"path": test_dir}})

            task = DirectoryRecreator(
                project_config=mock.Mock(),
                task_config=task_config,
                org_config=None,
            )

            assert isinstance(task.parsed_options.path, Path)
            assert str(task.parsed_options.path) == test_dir

    def test_path_as_path_object(self):
        """Test initialization with path as Path object."""
        with tempfile.TemporaryDirectory() as tmpdir:
            test_dir = Path(tmpdir) / "test_dir"
            task_config = TaskConfig({"options": {"path": test_dir}})

            task = DirectoryRecreator(
                project_config=mock.Mock(),
                task_config=task_config,
                org_config=None,
            )

            assert task.parsed_options.path == test_dir


class TestDirectoryRecreatorInitialization:
    """Test cases for DirectoryRecreator initialization methods."""

    def test_init_options_raises_error_when_path_is_file(self):
        """Test that _init_options raises TaskOptionsError when path is a file."""
        with tempfile.TemporaryDirectory() as tmpdir:
            # Create a file, not a directory
            test_file = os.path.join(tmpdir, "test_file.txt")
            with open(test_file, "w") as f:
                f.write("test content")

            task_config = TaskConfig({"options": {"path": test_file}})

            with pytest.raises(TaskOptionsError) as exc_info:
                DirectoryRecreator(
                    project_config=mock.Mock(),
                    task_config=task_config,
                    org_config=None,
                )

            assert f"Path {test_file} is a file" in str(exc_info.value)

    def test_init_options_succeeds_when_path_is_directory(self):
        """Test that _init_options succeeds when path is an existing directory."""
        with tempfile.TemporaryDirectory() as tmpdir:
            test_dir = os.path.join(tmpdir, "test_directory")
            os.makedirs(test_dir)

            task_config = TaskConfig({"options": {"path": test_dir}})

            task = DirectoryRecreator(
                project_config=mock.Mock(),
                task_config=task_config,
                org_config=None,
            )

            assert task.parsed_options.path == Path(test_dir)

    def test_init_options_succeeds_when_path_does_not_exist(self):
        """Test that _init_options succeeds when path doesn't exist yet."""
        with tempfile.TemporaryDirectory() as tmpdir:
            test_dir = os.path.join(tmpdir, "non_existent_directory")

            task_config = TaskConfig({"options": {"path": test_dir}})

            task = DirectoryRecreator(
                project_config=mock.Mock(),
                task_config=task_config,
                org_config=None,
            )

            assert task.parsed_options.path == Path(test_dir)


class TestDirectoryRecreatorRunTask:
    """Test cases for _run_task method."""

    def test_run_task_creates_new_directory(self):
        """Test _run_task creates a new directory when path doesn't exist."""
        with tempfile.TemporaryDirectory() as tmpdir:
            test_dir = os.path.join(tmpdir, "new_directory")

            task_config = TaskConfig({"options": {"path": test_dir}})

            project_config = mock.Mock()
            project_config.repo_root = tmpdir

            task = DirectoryRecreator(
                project_config=project_config,
                task_config=task_config,
                org_config=None,
            )

            # Directory should not exist yet
            assert not os.path.exists(test_dir)

            # Run the task
            task()

            # Directory should now exist
            assert os.path.exists(test_dir)
            assert os.path.isdir(test_dir)

    def test_run_task_recreates_existing_directory(self):
        """Test _run_task removes and recreates an existing directory."""
        with tempfile.TemporaryDirectory() as tmpdir:
            test_dir = os.path.join(tmpdir, "existing_directory")

            # Create directory with a file inside
            os.makedirs(test_dir)
            test_file = os.path.join(test_dir, "test_file.txt")
            with open(test_file, "w") as f:
                f.write("test content")

            # Verify directory and file exist
            assert os.path.exists(test_dir)
            assert os.path.exists(test_file)

            task_config = TaskConfig({"options": {"path": test_dir}})

            project_config = mock.Mock()
            project_config.repo_root = tmpdir

            task = DirectoryRecreator(
                project_config=project_config,
                task_config=task_config,
                org_config=None,
            )

            # Run the task
            task()

            # Directory should exist but file should be gone
            assert os.path.exists(test_dir)
            assert os.path.isdir(test_dir)
            assert not os.path.exists(test_file)

    def test_run_task_logs_created_message_for_new_directory(self):
        """Test _run_task logs 'created' message for new directory."""
        with tempfile.TemporaryDirectory() as tmpdir:
            test_dir = os.path.join(tmpdir, "new_directory")

            task_config = TaskConfig({"options": {"path": test_dir}})

            project_config = mock.Mock()
            project_config.repo_root = tmpdir

            task = DirectoryRecreator(
                project_config=project_config,
                task_config=task_config,
                org_config=None,
            )

            with mock.patch.object(task.logger, "info") as mock_logger:
                task()

                # Check that the correct message was logged (may be called multiple times by BaseTask)
                log_messages = [call[0][0] for call in mock_logger.call_args_list]
                assert any(
                    "created" in msg
                    and "removed and created" not in msg
                    and str(test_dir) in msg
                    for msg in log_messages
                )

    def test_run_task_logs_removed_and_created_message_for_existing_directory(self):
        """Test _run_task logs 'removed and created' message for existing directory."""
        with tempfile.TemporaryDirectory() as tmpdir:
            test_dir = os.path.join(tmpdir, "existing_directory")
            os.makedirs(test_dir)

            task_config = TaskConfig({"options": {"path": test_dir}})

            project_config = mock.Mock()
            project_config.repo_root = tmpdir

            task = DirectoryRecreator(
                project_config=project_config,
                task_config=task_config,
                org_config=None,
            )

            with mock.patch.object(task.logger, "info") as mock_logger:
                task()

                # Check that the correct message was logged (may be called multiple times by BaseTask)
                log_messages = [call[0][0] for call in mock_logger.call_args_list]
                assert any(
                    "removed and created" in msg and str(test_dir) in msg
                    for msg in log_messages
                )

    def test_run_task_handles_nested_directory_path(self):
        """Test _run_task can create nested directory paths."""
        with tempfile.TemporaryDirectory() as tmpdir:
            test_dir = os.path.join(tmpdir, "parent", "child", "grandchild")

            task_config = TaskConfig({"options": {"path": test_dir}})

            project_config = mock.Mock()
            project_config.repo_root = tmpdir

            task = DirectoryRecreator(
                project_config=project_config,
                task_config=task_config,
                org_config=None,
            )

            # Run the task
            task()

            # All nested directories should exist
            assert os.path.exists(test_dir)
            assert os.path.isdir(test_dir)

    def test_run_task_preserves_parent_directories(self):
        """Test _run_task only removes the specified directory, not parents."""
        with tempfile.TemporaryDirectory() as tmpdir:
            parent_dir = os.path.join(tmpdir, "parent")
            child_dir = os.path.join(parent_dir, "child")

            # Create parent and child directories
            os.makedirs(child_dir)

            # Create files in both directories
            parent_file = os.path.join(parent_dir, "parent.txt")
            child_file = os.path.join(child_dir, "child.txt")

            with open(parent_file, "w") as f:
                f.write("parent content")
            with open(child_file, "w") as f:
                f.write("child content")

            task_config = TaskConfig({"options": {"path": child_dir}})

            project_config = mock.Mock()
            project_config.repo_root = tmpdir

            task = DirectoryRecreator(
                project_config=project_config,
                task_config=task_config,
                org_config=None,
            )

            # Run the task
            task()

            # Parent directory and its file should still exist
            assert os.path.exists(parent_dir)
            assert os.path.exists(parent_file)

            # Child directory should exist but its file should be gone
            assert os.path.exists(child_dir)
            assert not os.path.exists(child_file)

    def test_run_task_with_relative_path(self):
        """Test _run_task works with relative paths."""
        with tempfile.TemporaryDirectory() as tmpdir:
            original_cwd = os.getcwd()
            try:
                os.chdir(tmpdir)
                test_dir = "relative_test_dir"

                task_config = TaskConfig({"options": {"path": test_dir}})

                project_config = mock.Mock()
                project_config.repo_root = tmpdir

                task = DirectoryRecreator(
                    project_config=project_config,
                    task_config=task_config,
                    org_config=None,
                )

                # Run the task
                task()

                # Directory should exist
                assert os.path.exists(test_dir)
                assert os.path.isdir(test_dir)
            finally:
                os.chdir(original_cwd)

    def test_run_task_with_directory_containing_subdirectories(self):
        """Test _run_task removes directory with complex structure."""
        with tempfile.TemporaryDirectory() as tmpdir:
            test_dir = os.path.join(tmpdir, "complex_directory")

            # Create complex directory structure
            os.makedirs(os.path.join(test_dir, "subdir1", "subdir2"))
            os.makedirs(os.path.join(test_dir, "subdir3"))

            # Create files in various locations
            with open(os.path.join(test_dir, "file1.txt"), "w") as f:
                f.write("content1")
            with open(os.path.join(test_dir, "subdir1", "file2.txt"), "w") as f:
                f.write("content2")
            with open(
                os.path.join(test_dir, "subdir1", "subdir2", "file3.txt"), "w"
            ) as f:
                f.write("content3")

            task_config = TaskConfig({"options": {"path": test_dir}})

            project_config = mock.Mock()
            project_config.repo_root = tmpdir

            task = DirectoryRecreator(
                project_config=project_config,
                task_config=task_config,
                org_config=None,
            )

            # Run the task
            task()

            # Directory should exist but be empty
            assert os.path.exists(test_dir)
            assert os.path.isdir(test_dir)
            assert len(os.listdir(test_dir)) == 0


class TestDirectoryRecreatorEdgeCases:
    """Test cases for edge cases and special scenarios."""

    def test_task_with_path_containing_spaces(self):
        """Test task handles paths with spaces correctly."""
        with tempfile.TemporaryDirectory() as tmpdir:
            test_dir = os.path.join(tmpdir, "directory with spaces")

            task_config = TaskConfig({"options": {"path": test_dir}})

            project_config = mock.Mock()
            project_config.repo_root = tmpdir

            task = DirectoryRecreator(
                project_config=project_config,
                task_config=task_config,
                org_config=None,
            )

            task()

            assert os.path.exists(test_dir)
            assert os.path.isdir(test_dir)

    def test_task_with_path_containing_special_characters(self):
        """Test task handles paths with special characters."""
        with tempfile.TemporaryDirectory() as tmpdir:
            test_dir = os.path.join(tmpdir, "test-dir_123")

            task_config = TaskConfig({"options": {"path": test_dir}})

            project_config = mock.Mock()
            project_config.repo_root = tmpdir

            task = DirectoryRecreator(
                project_config=project_config,
                task_config=task_config,
                org_config=None,
            )

            task()

            assert os.path.exists(test_dir)
            assert os.path.isdir(test_dir)

    def test_task_called_multiple_times(self):
        """Test task can be called multiple times successfully."""
        with tempfile.TemporaryDirectory() as tmpdir:
            test_dir = os.path.join(tmpdir, "test_directory")

            task_config = TaskConfig({"options": {"path": test_dir}})

            project_config = mock.Mock()
            project_config.repo_root = tmpdir

            task = DirectoryRecreator(
                project_config=project_config,
                task_config=task_config,
                org_config=None,
            )

            # Run the task multiple times
            task()
            assert os.path.exists(test_dir)

            # Add a file
            with open(os.path.join(test_dir, "file.txt"), "w") as f:
                f.write("content")

            task()
            assert os.path.exists(test_dir)
            assert len(os.listdir(test_dir)) == 0

            task()
            assert os.path.exists(test_dir)
            assert len(os.listdir(test_dir)) == 0



from cumulusci.core.config.util import get_devhub_config
from cumulusci.core.exceptions import SalesforceDXException
from cumulusci.core.tasks import BaseTask
from cumulusci.core.versions import PackageType, PackageVersionNumber
from cumulusci.salesforce_api.utils import get_simple_salesforce_connection
from cumulusci.utils.options import CCIOptions, Field


class GetPackageVersion(BaseTask):
    """Custom task to get package version ID"""

    class Options(CCIOptions):
        package_name: str = Field(
            ..., description="Package name to get package version ID for."
        )
        package_version: str = Field(
            ..., description="Package version to get package version ID for."
        )
        prefix: str = Field(
            "",
            description="Prefix to add to the package name.",
        )
        suffix: str = Field(
            "",
            description="Suffix to add to the package name.",
        )
        fail_on_error: bool = Field(
            False, description="Fail on error. [default to False]"
        )

    parsed_options: Options

    def _init_options(self, kwargs):
        super()._init_options(kwargs)
        self.parsed_options.package_version = PackageVersionNumber.parse(
            self.parsed_options.package_version, package_type=PackageType.SECOND_GEN
        )

    def _init_task(self):
        self.tooling = get_simple_salesforce_connection(
            self.project_config,
            get_devhub_config(self.project_config),
            api_version=self.project_config.project__package__api_version,
            base_url="tooling",
        )

    def _run_task(self):
        package_name = f"{self.parsed_options.prefix}{self.parsed_options.package_name}{self.parsed_options.suffix}".strip()

        query = (
            f"SELECT Id, SubscriberPackageVersionId FROM Package2Version WHERE Package2.Name='{package_name}' AND "
            f"MajorVersion={self.parsed_options.package_version.MajorVersion} AND "
            f"MinorVersion={self.parsed_options.package_version.MinorVersion} AND "
            f"PatchVersion={self.parsed_options.package_version.PatchVersion} AND "
            f"BuildNumber={self.parsed_options.package_version.BuildNumber}"
        )

        res = self.tooling.query(query)
        if res["size"] == 0:
            msg = f"Package version {package_name} {self.parsed_options.package_version} not found"
            self.logger.warning(msg)
            if self.parsed_options.fail_on_error:
                raise SalesforceDXException(msg)
            return

        if res["size"] > 1:
            msg = f"Multiple package versions found for {package_name} {self.parsed_options.package_version}"
            self.logger.warning(msg)
            if self.parsed_options.fail_on_error:
                raise SalesforceDXException(msg)

        self.return_values["package_version_id"] = res["records"][0]["Id"]
        self.return_values["subscriber_package_version_id"] = res["records"][0][
            "SubscriberPackageVersionId"
        ]
        self.return_values["package_name"] = package_name
        self.return_values["package_version"] = self.parsed_options.package_version

        self.logger.info(
            f"Package version {package_name} {self.parsed_options.package_version} found"
        )
        self.logger.info(
            f"Package version id: {self.return_values['package_version_id']}"
        )
        self.logger.info(
            f"SubscriberPackageVersion Id: {self.return_values['subscriber_package_version_id']}"
        )

        return self.return_values



from cumulusci.tasks.bulkdata.base_generate_data_task import BaseGenerateDataTask


class GenerateDummyData(BaseGenerateDataTask):
    """Generate data based on test mapping.yml"""

    def _read_mappings(self, mapping_file_path):
        assert mapping_file_path is None
        return None

    def generate_data(*args, **kwargs):
        return None
